import multiprocessing
import pickle
import shutil
import subprocess
import threading
from difflib import SequenceMatcher
from enum import Enum
from pathlib import Path
from typing import Union

import sys
import time
import zmq
from PyQt5.QtGui import QFont
from qtpy.QtCore import QLockFile
from qtpy.QtCore import QObject
from qtpy.QtCore import QThread
from qtpy.QtCore import Qt
from qtpy.QtCore import Signal
from qtpy.QtGui import QCloseEvent
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication
from qtpy.QtWidgets import QFrame
from qtpy.QtWidgets import QGridLayout
from qtpy.QtWidgets import QHBoxLayout
from qtpy.QtWidgets import QLabel
from qtpy.QtWidgets import QMainWindow
from qtpy.QtWidgets import QMessageBox
from qtpy.QtWidgets import QScrollArea
from qtpy.QtWidgets import QVBoxLayout, QGroupBox
from zmq.backend.cython._zmq import ZMQError

from egse.confman import ConfigurationManagerProxy, is_configuration_manager_active
from egse.gui import show_info_message
from egse.gui.buttons import ToggleButton, TouchButton
from egse.gui.led import LED, Indic
from egse.observer import Observer, Observable
from egse.plugin import entry_points
from egse.process import ProcessStatus
from egse.procman import LOGGER, StartCommand, StopCommand, StatusCommand
from egse.procman import ProcessManagerProxy
from egse.registry.client import RegistryClient
from egse.resource import get_resource
from egse.setup import Setup
from egse.system import do_every
from egse.zmq_ser import set_address_port

UPDATE_INTERVAL = 30

DEVICE_CMD_ENTRY_POINT = "cgse.service.device_command"
GUI_SCRIPTS_ENTRY_POINT = "gui_scripts"


class ControlServerStatus(Enum):
    """Status of the Control Server of a device."""

    ACTIVE = True  # Control Server active -> LED green
    INACTIVE = False  # Control Server inactive -> LED red
    UNKNOWN = None  # Control Server starting/stopping -> Hourglass


def get_cgse_cmd(device_proxy: str) -> str:
    """Determines the CGSE command for the Control Server of the given device proxy.

    To start/stop any device Control Server or query its status, we need to know its commanding.  These commands
    would be:

        - cgse <service_name> start <device_id> [<device_args>] [--sim]
        - cgse <service_name> stop <device_id>
        - cgse <service_name> status <device_id>

    In this function, we determine - based on the given device proxy - what the corresponding `service_name` is in the
    commands listed above.

    The current implementation (which may have to be changed in the future) is to compare the module name of the given
    device proxy with all `cgse.service` entry points in the `pyproject.toml` files of the installed modules.  The
    entry point of which the value matches the module name best is then selected.

    Args:
        device_proxy (str): Device proxy

    Returns: CGSE command to start/stop the Control Server or to query its status.
    """

    module_name = device_proxy[7:].rsplit(".", 1)[0]
    entry_point_values = []
    for ep in sorted(entry_points(DEVICE_CMD_ENTRY_POINT), key=lambda x: x.name):
        entry_point_values.append(ep.value)

    similarity_scores = [
        SequenceMatcher(None, entry_point_value, module_name).ratio() for entry_point_value in entry_point_values
    ]
    best_match_index = similarity_scores.index(max(similarity_scores))
    best_match = entry_point_values[best_match_index].split(":")[-1]

    return f"cgse {best_match}"


def get_cgse_ui(device_proxy: str) -> Union[str, None]:
    """Determines the CGSE command for the UI of the given device proxy.

    To open the UI of any device Control Server, we need to know its commanding. In this function, we determine - based
    on the given proxy - what the corresponding UI command is.

    The current implementation (which may have to be changed in the future) is to compare the module name of the given
    device proxy with all `gui_scripts` entries in the `pyproject.toml` files of the installed modules.  The entry of
    which the name matches the module name best in then selected (when its similarity score is high enough).

    Args:
        device_proxy (str): Device proxy

    Returns: CGSE command to start/stop the UI of the Control Server.
    """

    module_name = device_proxy[7:].rsplit(".", 1)[0]
    entry_point_values = []
    for ep in sorted(entry_points(DEVICE_CMD_ENTRY_POINT), key=lambda x: x.name):
        entry_point_values.append(ep.value)

    similarity_scores = [
        SequenceMatcher(None, entry_point_value, module_name).ratio() for entry_point_value in entry_point_values
    ]
    best_match_index = similarity_scores.index(max(similarity_scores))
    best_match = entry_point_values[best_match_index].split(":")[-1]

    return f"{best_match}_ui"

    # module_name = device_proxy[7:].rsplit(".", 1)[0]
    # entry_point_values = []
    # for ep in sorted(entry_points(GUI_SCRIPTS_ENTRY_POINT), key=lambda x: x.name):
    #     entry_point_values.append(ep.name)
    #
    # similarity_scores = [
    #     SequenceMatcher(None, entry_point_value, module_name).ratio() for entry_point_value in entry_point_values
    # ]
    # best_match_index = similarity_scores.index(max(similarity_scores))
    #
    # if similarity_scores[best_match_index] > 1 / len(similarity_scores):
    #     best_match = entry_point_values[best_match_index].split(":")[-1]
    #     return best_match
    # else:
    #     return None


class UiCommand:
    """Command to start the UI for the Control Server for a device."""

    def __init__(self, device_id: str, cmd: str):
        """Initialisation of a UI command.

        Args:
            device_id (str): Device identifier
            cmd (str): Command to open the UI for the Control Server
        """

        self._device_id = device_id
        self._cmd = cmd

    @property
    def device_id(self):
        """Returns the device identifier.

        Returns: Device identifier
        """
        return self._device_id

    @property
    def cmd(self) -> str:
        """Returns the full command to open the UI for the Control Server.

        Returns: Full command to open the UI for the Control Server.
        """
        return self._cmd


class ConfigurationManagerMonitoringWorker(QObject):
    """Monitoring worker for the Configuration Manager."""

    setup_changed_signal = Signal(Setup)
    obsid_changed_signal = Signal(int)

    def __init__(self, parent=None):
        """Initialisation of a monitoring worker for the Configuration Manager.

        This monitoring worker will listen on the monitoring port of the Configuration Manager and send out a signal
        whenever the setup or obsid is changed.
        """

        super(ConfigurationManagerMonitoringWorker, self).__init__(parent)

        self.setup = None  # Previous setup
        self.obsid = None  # Previous obsid

        self.active = False

        self.socket = zmq.Context().socket(zmq.SUB)
        self.connect_socket()

        self.monitoring_info = None

    def connect_socket(self) -> None:
        """Connect the socket to the monitoring port of the Configuration Manager."""

        cm = ConfigurationManagerProxy()

        endpoint = cm.get_endpoint()
        monitoring_port = cm.get_monitoring_port()
        address = set_address_port(endpoint, monitoring_port)

        self.socket.connect(address)
        self.socket.setsockopt_string(zmq.SUBSCRIBE, "")

    def start_listening(self) -> None:
        """Start listening to the monitoring port of the Configuration Manager."""

        self.active = True
        self.run()

    def stop_listening(self) -> None:
        """Stop listening to the monitoring port of the Configuration Manager."""

        self.active = False

    def run(self) -> None:
        """Keep on listening on the monitoring port as long as the monitoring worker is active.

        Compare the setup and obsid from the incoming monitoring information from the Configuration Manager with the
        last knows values.  Whenever the setup or obsid has changed, a signal is emitted.
        """

        while self.active:
            pickle_string = self.socket.recv()
            self.monitoring_info = pickle.loads(pickle_string)

            self.check_for_setup_changes()
            self.check_for_obsid_changes()

            time.sleep(UPDATE_INTERVAL)

    def check_for_setup_changes(self) -> None:
        """Checks the incoming monitoring information from the Configuration Manager for changes in the setup.

        If the setup has changed, a signal is emitted.  A change in setup is defined as a change in the setup identifier.
        """

        new_setup = self.monitoring_info["setup"]

        if new_setup != self.setup:
            self.setup = new_setup

            # Due to static type checking the IDE doesn't recognise the `emit` method on a `Signal` object.  This
            # happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a way that static
            # analysers can easily detect.
            # noinspection PyUnresolvedReferences
            self.setup_changed_signal.emit(self.setup)

    def check_for_obsid_changes(self) -> None:
        """Checks the incoming monitoring information from the Configuration Manager for changes in the obsid.

        If the obsid has changes, a signal is emitted.
        """

        new_obsid = self.monitoring_info["obsid"]

        if new_obsid != self.obsid:
            self.obsid = new_obsid

            # Due to static type checking the IDE doesn't recognise the `emit` method on a `Signal` object.  This
            # happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a way that static
            # analysers can easily detect.
            # noinspection PyUnresolvedReferences
            self.obsid_changed_signal.emit(self.obsid)


class CoreServiceMonitoringWorker(QObject):
    """Monitoring worker for the core services."""

    core_service_status_signal = Signal(dict)

    def __init__(self, core_service_name: str, core_service: str, parent=None):
        """Initialisation of a monitoring worker for a core service.

        Args:
            core_service_name (str): Name of the core service
            core_service (str): Core service command to start / stop / query the status with
                                `cgse <core_service> start|stop|status`
        """

        super().__init__(parent)

        self.core_service_name = core_service_name
        self.cs = self.service_type = core_service

        self.active = False
        self.registry_client = RegistryClient()

        self.core_service_is_active = False

    def start_listening(self):
        """Start checking the status of the core service as long as the monitoring worker is active."""

        self.active = True
        self.registry_client.connect()
        self.run()

    def stop_listening(self):
        """Stop checking the status of the core service."""

        self.active = False
        self.registry_client.disconnect()

    def run(self):
        """Keep on checking the status of the core service as long as the monitoring worker is active.

        Whenever the status of the core service has changed, a signal is emitted.  This signal comprises a
        dictionary with the following information:

            - Core service identifier;
            - Boolean indicating whether the core service is active;
        """

        while self.active:
            try:
                services = self.registry_client.list_services(service_type=self.service_type)

                # Due to static type checking the IDE doesn't recognise the `emit` method on a `Signal` object.
                # This happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a
                # way that static analysers can easily detect.
                # noinspection PyUnresolvedReferences
                self.core_service_status_signal.emit(
                    {"core_service_name": self.core_service_name, "core_service_is_active": len(services) > 0}
                )

                time.sleep(UPDATE_INTERVAL)
            except ZMQError:
                pass


class DeviceMonitoringWorker(QObject):
    """Monitoring worker for the devices."""

    process_status_signal = Signal(dict)

    def __init__(self, device_id: str, device_args: list, device_proxy: str, parent=None):
        """Initialisation of a monitoring worker for a device.

        Args:
            device_id (str): Identifier of the device to monitor.
            device_args (list): Arguments for the device.
            device_proxy (str): Device proxy.
        """

        super().__init__(parent)

        self.device_id = device_id
        self.device_args = device_args
        self.device_proxy = device_proxy
        self.process_manager: ProcessManagerProxy = ProcessManagerProxy()
        self.active = False
        self.registry_client = RegistryClient()

        self.cs_is_active = False
        self.cs_status = ControlServerStatus.INACTIVE

        self.cgse_cmd = get_cgse_cmd(self.device_proxy)
        self.status_cmd = StatusCommand(device_id=self.device_id, device_args=self.device_args, cgse_cmd=self.cgse_cmd)

    def start_listening(self) -> None:
        """Start listening to the output of the CGSE status command."""

        self.active = True
        self.registry_client.connect()
        self.run()

    def stop_listening(self) -> None:
        """Stop listening to the output of the CGSE status command."""

        self.active = False
        self.registry_client.disconnect()

    def run(self) -> None:
        """Keep on checking the status of the device Control Server as long as the monitoring worker is active.

        Execute the CGSE status command to query the status of the device Control Server.  Whenever the status of the
        device Control Server has changed, a signal is emitted.  This signal comprises a dictionary with the following
        information:
            - Device identifier;
            - Boolean indicating whether the Control Server is active;
            - In case the Control Server is active:
                - Whether the device is connected;
                - Whether the Control Server was started in simulator mode (rather than operational mode).
        """

        while self.active:
            try:
                services = self.registry_client.list_services(service_type=self.device_id.upper())

                if len(services) == 0:
                    cs_is_active_new = False
                    if cs_is_active_new != self.cs_is_active:
                        self.cs_is_active = cs_is_active_new
                        self.process_status_signal.emit({"device_id": self.device_id, "cs_is_active": False})

                else:
                    cs_is_active_new = not services is None
                    if cs_is_active_new != self.cs_is_active:
                        self.cs_is_active = cs_is_active_new

                        if self.cs_is_active:
                            self.cs_status = ControlServerStatus.ACTIVE
                            status_output = self.process_manager.get_device_process_status(self.status_cmd)

                            # Due to static type checking the IDE doesn't recognise the `emit` method on a `Signal` object.
                            # This happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a
                            # way that static analysers can easily detect.
                            # noinspection PyUnresolvedReferences
                            self.process_status_signal.emit(status_output)
            except ZMQError:
                pass


class LedColor(Enum):
    """Potential colours for the LEDs showing the status of the processes."""

    ACTIVE = Indic.GREEN  # Not for device Control Servers
    ACTIVE_DEVICE_CONNECTED = Indic.GREEN  # Device Control Servers only
    ACTIVE_NO_DEVICE_CONNECTED = Indic.ORANGE  # Device Control Servers only
    INACTIVE = Indic.RED


class CoreServiceWidget(QGroupBox):
    """Widget to display the status of the core services."""

    def __init__(self, core_service_name: str, parent: QMainWindow):
        """Initialisation of a widget to display the status of the core services.

        Args:
            core_service_name (str): Name of the core service
            parent (QMainWindow): Parent window
        """

        super().__init__()

        self.core_service_name = core_service_name

        self.parent = parent

        layout = QGridLayout()

        index = 0

        # Status LED

        self.status_led = LED(parent=self)
        layout.addWidget(self.status_led, 0, index)
        index += 1

        # Process name

        self.core_service_name_label = QLabel(core_service_name)
        layout.addWidget(self.core_service_name_label, 0, index)
        layout.setColumnStretch(index, 1)  # Push LED and name to the left and buttons to the right
        index += 1

        self.setLayout(layout)

    def set_status_led(self, color: LedColor) -> None:
        """Set the colour of the status LED.

        Args:
            color (LedColor): LED colour
        """

        self.status_led.set_color(color.value)


class DeviceWidget(QGroupBox, Observable):
    """Widget to display the status of a device Control Server, to start and stop it, and to open its UI."""

    def __init__(
        self, device_id: str, device_name: str, device_proxy: str, device_args: Union[list, None], parent: QMainWindow
    ):  # FIXME device_args
        """Initialisation of a device widget.

        Args:
            device_id (str): Device identifier
            device_name (str): Device name
            device_proxy (str): Device proxy
            device_args (Union[str, args]): Device arguments
            parent (QMainWindow): Parent window
        """

        super().__init__()
        Observable.__init__(self)

        self.device_id = device_id
        self.device_args = device_args
        self.is_simulator_mode = False

        self.parent = parent

        self.cgse_cmd = get_cgse_cmd(device_proxy)
        self.ui_cmd = get_cgse_ui(device_proxy)

        layout = QGridLayout()

        index = 0

        # Status LED

        self.status_led = LED(parent=self)
        self.set_status_led(LedColor.INACTIVE)
        layout.addWidget(self.status_led, 0, index)
        index += 1

        # Process name

        self.process_name_label = QLabel(device_name)
        layout.addWidget(self.process_name_label, 0, index)
        layout.setColumnStretch(index, 1)  # Push LED and name to the left and buttons to the right
        index += 1

        # Open UI
        if self.has_ui_option():
            self.ui_button = TouchButton(
                name=f"Open the UI for the {self.device_id} Control Server.",
                status_tip=f"Open the UI for the {self.device_id} Control Server.",
                selected=get_resource(":/icons/user-interface.svg"),
            )
            self.ui_button.setFixedSize(30, 30)
            self.ui_button.clicked.connect(self.open_ui)
            layout.addWidget(self.ui_button, 0, index)
            index += 1

        # Shut down / re-start

        self.start_stop_button = ToggleButton(
            name=f"Start / shut down the {self.device_id} Control Server.",
            status_tip=f"Start / shut down the {self.device_id} Control Server.",
            selected=get_resource(":/icons/start-process-button.svg"),
            not_selected=get_resource(":/icons/stop-process-button.svg"),
            disabled=[get_resource(":/icons/busy.svg"), get_resource(":/icons/busy.svg")],
        )
        self.start_stop_button.clicked.connect(self.start_stop_cs)
        layout.addWidget(self.start_stop_button, 0, index)
        index += 1

        # Operational / simulator mode

        if self.has_sim_option():
            self.simulator_option_button = ToggleButton(
                name=f"Operational vs. simulator mode",
                status_tip=f"Indicate whether you want to start the Control Server in operational or simulator mode.",
                selected=get_resource(":/icons/operational-mode.svg"),
                not_selected=get_resource(":/icons/simulator-mode.svg"),
                disabled=[get_resource(":/icons/simulator-mode.svg"), get_resource(":/icons/operational-mode.svg")],
            )
        else:
            self.simulator_option_button = ToggleButton(
                name=f"Operational mode only",
                status_tip=f"Operational mode only.",
                selected=get_resource(":/icons/operational-mode.svg"),
                not_selected=get_resource(":/icons/operational-mode.svg"),
                disabled=[get_resource(":/icons/operational-mode.svg")],
            )
        self.simulator_option_button.clicked.connect(self.change_cs_start_mode)
        layout.addWidget(self.simulator_option_button, 0, index)
        index += 1

        self.setLayout(layout)

    def has_sim_option(self) -> bool:
        """Checks whether the Control Server can be started in simulator mode.

        Returns: True if the Control Server can be started in simulator mode; False otherwise.
        """

        output = subprocess.check_output(f"{self.cgse_cmd} start --help", shell=True).decode("utf-8")
        return "--sim" in output

    def has_ui_option(self) -> bool:
        """Checks whether the Control Server has a corresponding UI.

        Returns: True if the Control Server has a corresponding UI.
        """

        if not self.ui_cmd:
            return False
        else:
            return shutil.which(self.ui_cmd) is not None

        # return self.ui_cmd is not None and

    def open_ui(self) -> None:
        """Open the UI."""

        self.notify_observers(UiCommand(device_id=self.device_id, cmd=self.ui_cmd))

    def start_stop_cs(self) -> None:
        """Take action when the start/stop button is clicked."""

        self.start_stop_button.disable()

        if self.start_stop_button.is_selected():
            self.notify_observers(
                StopCommand(device_id=self.device_id, device_args=self.device_args, cgse_cmd=self.cgse_cmd)
            )
        else:
            self.notify_observers(
                StartCommand(
                    device_id=self.device_id,
                    cgse_cmd=self.cgse_cmd,
                    device_args=self.device_args,
                    is_simulator_mode=self.is_simulator_mode,
                )
            )

    def change_cs_start_mode(self) -> None:
        """Switch from simulator to operational mode, and vice versa."""

        self.is_simulator_mode = not self.is_simulator_mode

    def set_status_led(self, color: LedColor) -> None:
        """Set the colour of the status LED.

        Args:
            color (LedColor): LED colour
        """

        self.status_led.set_color(color.value)


class ProcessManagerUIModel:
    """Model in the MVC pattern that makes the PM UI application."""

    def __init__(self):
        """Initialisation of the Process Manager UI model."""

        super().__init__()

        try:
            self.process_manager: Union[ProcessManagerProxy, None] = ProcessManagerProxy()
        except ConnectionError:
            LOGGER.error("Could not connect to Process Manager Control Server")
            self.process_manager: Union[ProcessManagerProxy, None] = None

    def is_connected(self) -> bool:
        """Checks whether the Process Manager Control Server is active.

        Checks whether a connection to the Process Manager Control Server has been established.

        Returns: True if a connection to the Process Manager Control Server has been established; False otherwise.
        """

        if not self.process_manager:
            return False

        return self.process_manager.ping() and self.process_manager.is_cs_connected()

    def get_device_ids(self) -> dict:
        """Returns the relevant device identifiers.

        Returns: Dictionary in which the device identifiers are the keys, and the values are a tuple with the device
                 name, the device proxy and the device arguments.
        """

        return self.process_manager.get_device_ids()

    def get_core_services(self) -> dict:
        """Returns the core services.

        Returns: Dictionary with the core service names are the keys, and the values are the core service CGSE commands.
        """

        return self.process_manager.get_core_processes()

    def start_process(self, start_cmd: StartCommand):
        """Start a process with the given start command.

        The process is started on the same machine as the Process Manager.

        Args:
            start_cmd (StartCommand): Command to start the process
        """
        self.process_manager.start_process(start_cmd)

    def stop_process(self, stop_cmd: StopCommand):
        """Stop a process with the given stop command.

        The process was running on the same machine as the Process Manager.

        Args:
            stop_cmd (StopCommand): Command to stop the process
        """

        self.process_manager.stop_process(stop_cmd)


class ProcessManagerUIView(QMainWindow, Observable):
    """View in the MVC pattern that makes the PM UI application."""

    def __init__(self, core_services: dict):
        """Initialisation of the Process Manager UI model with the given core services.

        Args:
            core_services (dict): Dictionary with the core services
        """

        super(ProcessManagerUIView, self).__init__()
        Observable.__init__(self)

        self.setGeometry(300, 300, 1000, 1000)
        self.setWindowTitle("Process Manager")

        self.setup_id = QLabel("No setup loaded")
        self.setup_id.setStyleSheet("QGroupBox { font-size: 18px;}")
        self.obsid = QLabel("No observation is running")
        self.sut_id = QLabel("SUT ID unknown")

        self.core_services: dict = core_services
        self.overview_core_services_widget_layout = QVBoxLayout()
        self.overview_core_services_widget = QGroupBox("Core Services", self)

        self.overview_core_services_widget.setStyleSheet("QGroupBox { font-size: 18px;}")

        self.core_service_widgets = {}  # Widgets for the core services
        self.device_widgets = {}  # Widgets for the devices

        self.overview_devices_widget_layout = QVBoxLayout()
        self.overview_devices_widget = QGroupBox("Devices", self)
        self.overview_devices_widget.setStyleSheet("QGroupBox { font-size: 18px;}")

        self.init_ui()

    def init_ui(self) -> None:
        """Creating the content of the UI."""

        app_frame = QFrame()
        app_frame.setObjectName("AppFrame")

        global_layout = QHBoxLayout()

        vbox_left = QVBoxLayout()
        vbox_right = QVBoxLayout()
        global_layout.addLayout(vbox_left)
        global_layout.addLayout(vbox_right)

        app_frame.setLayout(global_layout)
        scroll = QScrollArea()

        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)

        # Obsid, setup ID, SUT

        info_gridbox = QGridLayout()

        self.obsid = QLabel("No observation is running")
        self.setStyleSheet("QMainWindow{border: 3px solid green;}")
        obsid_label = QLabel("Obsid:")
        obsid_label.setFont(QFont("Arial", 16))
        info_gridbox.addWidget(obsid_label, 0, 0)
        info_gridbox.addWidget(self.obsid, 0, 1)

        self.setup_id = QLabel("No setup loaded")
        setup_id_label = QLabel("Setup ID:")
        setup_id_label.setFont(QFont("Arial", 16))
        info_gridbox.addWidget(setup_id_label, 1, 0)
        info_gridbox.addWidget(self.setup_id, 1, 1)

        self.sut_id = QLabel("SUT ID unknown")
        sut_id_label = QLabel("SUT ID:")
        sut_id_label.setFont(QFont("Arial", 16))
        info_gridbox.addWidget(sut_id_label, 2, 0)
        info_gridbox.addWidget(self.sut_id, 2, 1)

        vbox_left.addLayout(info_gridbox)
        vbox_left.addWidget(QLabel())

        # Core Services

        self.overview_core_services_widget_layout.addStretch()
        self.overview_core_services_widget.setLayout(self.overview_core_services_widget_layout)

        for core_service in self.core_services.keys():
            core_service_widget = CoreServiceWidget(core_service_name=core_service, parent=self)
            self.overview_core_services_widget_layout.addWidget(core_service_widget)
            self.core_service_widgets[core_service] = core_service_widget

        vbox_left.addWidget(self.overview_core_services_widget)

        # Devices

        self.overview_devices_widget_layout.addStretch()
        self.overview_devices_widget.setLayout(self.overview_devices_widget_layout)

        vbox_right.addWidget(self.overview_devices_widget)

        scroll.setWidget(app_frame)
        self.setCentralWidget(scroll)

    def clear_device_overview(self) -> None:
        """Clears the panel with the overview of the device Control Servers."""

        while self.overview_devices_widget_layout.count():
            widget = self.overview_devices_widget_layout.takeAt(0).widget()
            if widget:
                widget.deleteLater()

                if widget := self.overview_devices_widget_layout.takeAt(0).widget():
                    widget.deleteLater()
                self.device_widgets.clear()

    def closeEvent(self, close_event: QCloseEvent) -> None:
        """Takes action when the UI is closed."""

        self.notify_observers(close_event)


class ProcessManagerUIController(Observer):
    """Controller in the MVC pattern that makes the PM UI application."""

    def __init__(self, model: ProcessManagerUIModel, view: ProcessManagerUIView):
        """Initialisation of the Controller for the Process Manager GUI.

        Args:
            model (ProcessManagerUIModel): Process Manager UI model
            view (ProcessManagerUIView): Process Manager UI view
        """

        super().__init__()

        self.model = model
        self.view = view

        # Monitoring the Configuration Manager -> setup + obsid

        self.configuration_manager_monitoring_thread = QThread(self.view)
        self.configuration_manager_monitoring_worker = ConfigurationManagerMonitoringWorker()
        self.configuration_manager_monitoring_worker.moveToThread(self.configuration_manager_monitoring_thread)
        # Due to static type checking the IDE doesn't recognise the `connect` method on a `Signal` object.  This
        # happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a way that static
        # analysers can easily detect.
        # noinspection PyUnresolvedReferences
        self.configuration_manager_monitoring_worker.setup_changed_signal.connect(self.on_setup_changed_signal)
        # noinspection PyUnresolvedReferences
        self.configuration_manager_monitoring_thread.started.connect(
            self.configuration_manager_monitoring_worker.start_listening
        )
        self.configuration_manager_monitoring_thread.start()

        # Monitoring the Core Services

        self.core_services = self.model.get_core_services()
        self.core_service_monitoring_workers = {}
        self.core_service_monitoring_threads = {}

        self.start_core_service_monitoring()

        # Monitoring the devices
        # -> Start doing so when the setup changes

        self.device_ids = {}

        self.device_monitoring_threads = {}
        self.device_monitoring_workers = {}

    def on_setup_changed_signal(self, setup: Setup) -> None:
        """Reaction to a change in the setup.

        This method is called when a new setup is loaded in the Configuration Manager and executes the following steps:
            - Based on the old list of devices (from the previous setup):
                - Stop monitoring the devices;
                - Remove the device widgets from the layout;
            - Update the list of devices (based on the new setup);
            - Based on the updated list of devices:
                - Start monitoring the devices;
                - Add the device widgets to the layout.

        Args:
            setup (Setup): New setup that was loaded in the Configuration Manager
        """

        LOGGER.info(f"Updating the list of devices in the PM UI, based on setup {setup.get_id()}")

        # Stop monitoring the current devices (if any) -> based on the previous setup
        self.stop_device_monitoring()

        # Remove the current device widgets from the layout -> based on the previous setup
        self.view.clear_device_overview()

        # Update the list of devices, based on the new setup
        self.device_ids = self.model.get_device_ids()

        # Start monitoring the new devices -> based on the new setup
        self.start_device_monitoring()

        self.view.setup_id.setText(f"{setup.get_id()}")
        self.view.sut_id.setText(setup.sut.id)

    def on_device_status_signal(self, device_info: dict) -> None:
        """Reaction to a change in the status of a device.

        Args:
            device_info (dict): Information about a device
        """

        device_id = device_info["device_id"]
        cs_is_active = device_info["cs_is_active"]

        try:
            if cs_is_active:
                device_is_connected = device_info["device_is_connected"]
                is_simulator_mode = device_info["is_simulator_mode"]
                if device_is_connected:
                    self.view.device_widgets[device_id].set_status_led(LedColor.ACTIVE_DEVICE_CONNECTED)
                else:
                    self.view.device_widgets[device_id].set_status_led(LedColor.ACTIVE_NO_DEVICE_CONNECTED)
                self.view.device_widgets[device_id].simulator_option_button.set_selected(is_simulator_mode)
                self.view.device_widgets[device_id].simulator_option_button.disable()
                self.view.device_widgets[device_id].start_stop_button.set_selected(False)
                self.view.device_widgets[device_id].start_stop_button.enable()

            else:
                self.view.device_widgets[device_id].set_status_led(LedColor.INACTIVE)
                self.view.device_widgets[device_id].simulator_option_button.enable()
                self.view.device_widgets[device_id].simulator_option_button.set_selected(True)
                self.view.device_widgets[device_id].is_simulator_mode = False
                self.view.device_widgets[device_id].start_stop_button.set_selected(True)
                self.view.device_widgets[device_id].start_stop_button.enable()

            self.view.device_widgets[device_id].start_stop_button.enable()
        except KeyError:
            pass

    def on_core_service_status_signal(self, core_service_info: dict) -> None:
        """Reaction to a change in the status of a core service.

        Args:
            core_service_info (dict): Information about a core service
        """

        core_service = core_service_info["core_service_name"]
        cs_is_active = core_service_info["core_service_is_active"]

        try:
            if cs_is_active:
                self.view.core_service_widgets[core_service].set_status_led(LedColor.ACTIVE)
            else:
                self.view.core_service_widgets[core_service].set_status_led(LedColor.INACTIVE)
        except KeyError:
            pass

    def stop_configuration_manager_monitoring(self) -> None:
        """Stop monitoring the Configuration Manager."""

        self.configuration_manager_monitoring_worker.stop_listening()
        self.configuration_manager_monitoring_thread.quit()
        self.configuration_manager_monitoring_thread.wait()

    def start_core_service_monitoring(self):
        """Start monitoring the core services."""

        for core_service_name, core_service_cmd in self.model.get_core_services().items():
            core_service_monitoring_thread = QThread(self.view)
            core_service_monitoring_worker = CoreServiceMonitoringWorker(core_service_name, core_service_cmd)
            core_service_monitoring_worker.moveToThread(core_service_monitoring_thread)
            # Due to static type checking the IDE doesn't recognise the `connect` method on a `Signal` object.  This
            # happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a way that static
            # analysers can easily detect.
            # noinspection PyUnresolvedReferences
            core_service_monitoring_worker.core_service_status_signal.connect(self.on_core_service_status_signal)
            # noinspection PyUnresolvedReferences
            core_service_monitoring_thread.started.connect(core_service_monitoring_worker.start_listening)
            core_service_monitoring_thread.start()

            self.core_service_monitoring_threads[core_service_name] = core_service_monitoring_thread
            self.core_service_monitoring_workers[core_service_name] = core_service_monitoring_worker

    def stop_core_service_monitoring(self):
        """Stop monitoring the core services."""

        for core_service in self.model.get_core_services():
            self.core_service_monitoring_workers[core_service].stop_listening()
            self.core_service_monitoring_threads[core_service].quit()
            self.core_service_monitoring_threads[core_service].wait()

        self.core_service_monitoring_workers.clear()
        self.core_service_monitoring_threads.clear()

    def start_device_monitoring(self):
        """Start monitoring the devices and add them to the UI.

        This method is called when the setup changes.  We assume that the following steps have already been executed:
            - We stopped monitoring the devices that were listed in the previous setup:
            - We removed the corresponding device widgets from the layout;
            - We updated the list of devices to monitor (in `self.device_ids`) based on the new setup.

        This method does the following:
            - Start monitoring the devices that are listed in the new setup;
            - Add the new device widgets to the layout.
        """

        for device_id, (device_name, device_proxy, device_args) in self.device_ids.items():
            device_monitoring_thread = QThread(self.view)  # TODO
            device_monitoring_worker = DeviceMonitoringWorker(device_id, device_args, device_proxy)
            device_monitoring_worker.moveToThread(device_monitoring_thread)
            # Due to static type checking the IDE doesn't recognise the `connect` method on a `Signal` object.  This
            # happens because `Signal` is a special descriptor in `PyQt` and doesn't expose `emit` in a way that static
            # analysers can easily detect.
            # noinspection PyUnresolvedReferences
            device_monitoring_worker.process_status_signal.connect(self.on_device_status_signal)
            # noinspection PyUnresolvedReferences
            device_monitoring_thread.started.connect(device_monitoring_worker.start_listening)
            device_monitoring_thread.start()

            self.device_monitoring_threads[device_id] = device_monitoring_thread
            self.device_monitoring_workers[device_id] = device_monitoring_worker

            # Add the new device widgets

            device_widget = DeviceWidget(device_id, device_name, device_proxy, device_args, self.view)
            device_widget.add_observer(self)
            self.view.device_widgets[device_id] = device_widget
            self.view.overview_devices_widget_layout.addWidget(device_widget)

    def stop_device_monitoring(self):
        """Stop monitoring the devices and remove them from the UI."""

        if len(self.device_ids) > 0:
            for device_id in self.device_ids.keys():
                self.device_monitoring_workers[device_id].stop_listening()
                self.device_monitoring_threads[device_id].quit()
                self.device_monitoring_threads[device_id].wait()

        self.device_monitoring_workers.clear()
        self.device_monitoring_threads.clear()

    def update(self, changed_object) -> None:
        """Updates the state of the controller upon notification from an observable."""

        # Closure of the UI

        if isinstance(changed_object, QCloseEvent):
            # Stop monitoring the Configuration Manager
            self.stop_configuration_manager_monitoring()

            # Stop monitoring the Core Services
            self.stop_core_service_monitoring()

            # Stop monitoring the devices
            self.stop_device_monitoring()

        # Start button in one of the widgets has been clicked

        elif isinstance(changed_object, StartCommand):
            self.model.start_process(changed_object)
            self.device_monitoring_workers[changed_object.device_id].cs_status = ControlServerStatus.UNKNOWN

        # Stop button in one of the widgets has been clicked

        elif isinstance(changed_object, StopCommand):
            self.model.stop_process(changed_object)
            self.device_monitoring_workers[changed_object.device_id].cs_status = ControlServerStatus.UNKNOWN

        # UI button in one of the widgets has been clicked

        elif isinstance(changed_object, UiCommand):
            subprocess.call(changed_object.cmd, shell=True)

    def do(self, actions):
        """Execute the given actions upon notification from an observable."""

        pass


def main():
    """Main method to launch the Process Manager GUI."""

    lock_file = QLockFile(str(Path("~/pm_ui.app.lock").expanduser()))

    multiprocessing.current_process().name = "pm_ui"
    app = QApplication(sys.argv)

    app.setWindowIcon(QIcon(str(get_resource(":/icons/pm_ui.svg"))))

    if lock_file.tryLock(100):
        process_status = ProcessStatus()

        timer_thread = threading.Thread(target=do_every, args=(10, process_status.update))
        timer_thread.daemon = True
        timer_thread.start()

        # Check whether the Process Manager CS is running
        # (show a warning in a pop-up window if it's not)

        try:
            with ProcessManagerProxy():
                if not is_configuration_manager_active():
                    description = "Could not connect to Configuration Manager"
                    into_text = (
                        "The GUI will start, but without listed processes. "
                        "Please, check if the Configuration Manager is running and start the server if needed."
                        # "Otherwise, check if the correct HOSTNAME for the Configuration Manager is set in the "
                        # "Settings.yaml "
                        # "configuration file."
                        "The Process Manager GUI will have to be re-started after that."
                    )

                    show_info_message(description, into_text)

        except ConnectionError:
            description = "Could not connect to Process Manager Control Server"

            into_text = (
                "The GUI will start, but the connection button will show a disconnected state. "
                "Please, check if the Control Server is running and start the server if needed. "
                "Otherwise, check if the correct HOSTNAME for the control server is set in the "
                "Settings.yaml "
                "configuration file."
            )

            show_info_message(description, into_text)
            return

        # Create the Process Manager GUI, following the MVC-model

        model = ProcessManagerUIModel()
        core_services = model.get_core_services()
        view = ProcessManagerUIView(core_services)
        controller = ProcessManagerUIController(model, view)
        view.add_observer(controller)

        view.show()

        return app.exec_()
    else:
        error_message = QMessageBox()
        error_message.setIcon(QMessageBox.Warning)
        error_message.setWindowTitle("Error")
        error_message.setText("The Process Manager (PM) GUI application is already running!")
        error_message.setStandardButtons(QMessageBox.Ok)

        return error_message.exec()


if __name__ == "__main__":
    sys.exit(main())



from enum import Enum
from pathlib import Path

import numpy as np

from egse.settings import Settings

HERE = Path(__file__).parent

DEVICE_SETTINGS = Settings.load("Aim-TTi TGF4000")
CMD_DELAY = DEVICE_SETTINGS["CMD_DELAY"]
CS_SETTINGS = Settings.load("Aim-TTi TGF4000 Control Server")
PROTOCOL = CS_SETTINGS.get("PROTOCOL", "tcp")  # Communication protocol

PROXY_TIMEOUT = 10


class Version(float):
    """Version float of XX.xx format."""

    def __str__(self):
        return "{0:05.2f}".format(float(self))


# class ArbDataFile(str, Enum):
#     """Enumeration for the files with the ARB data (for the AWG).
#
#     The files are located in the /arbdata folder.
#     """
#
#     # Original Evoleo files to configure Clk_ccdread with short pulses of 200 ms
#
#     N_CCD_READ_25 = ("ccdRead25.arb",)  # CCD readout for N-CAM (image cycle time: 25s)
#     N_CCD_READ_31_25 = ("ccdRead31_25.arb",)  # CCD readout for N-CAM (image cycle time: 31.25s)
#     N_CCD_READ_37_50 = ("ccdRead37_50.arb",)  # CCD readout for N-CAM (image cycle time: 37.50s)
#     N_CCD_READ_43_75 = ("ccdRead43_75.arb",)  # CCD readout for N-CAM (image cycle time: 43.75s)
#     N_CCD_READ_50 = ("ccdRead50.arb",)  # CCD readout for N-CAM (image cycle time: 50s)
#
#     F_CCD_READ = ("FccdRead.arb",)  # CCD readout for F-CAM (fixed image cycle time)
#     F_CCD_READ_MIN_POINTS = "FccdRead_min_points.arb"  # Taken from PTO-EST-SC-TN-1563 (v1.0)
#
#     # Files modified by KUL to configure Clk_ccdread with short pulses of 150ms
#
#     N_CCD_READ_25_150MS = ("ccdRead25_150ms.arb",)  # CCD readout for N-CAM (image cycle time: 25s)
#     N_CCD_READ_31_25_150MS = ("ccdRead31_25_150ms.arb",)  # CCD readout for N-CAM (image cycle time: 31.25s)
#     N_CCD_READ_37_50_150MS = ("ccdRead37_50_150ms.arb",)  # CCD readout for N-CAM (image cycle time: 37.50s)
#     N_CCD_READ_43_75_150MS = ("ccdRead43_75_150ms.arb",)  # CCD readout for N-CAM (image cycle time: 43.75s)
#     N_CCD_READ_50_150MS = ("ccdRead50_150ms.arb",)  # CCD readout for N-CAM (image cycle time: 50s)
#
#     # Original Evoleo files to configure Clk_heater
#
#     SVM_SYNC_CCD_READ_25 = ("HeaterSync_ccdRead25.arb",)  # SVM/TCS sync signal (image cycle time: 25s)
#     SVM_SYNC_CCD_READ_31_25 = ("HeaterSync_ccdRead31_25.arb",)  # SVM/TCS sync signal (image cycle time: 31.25s)
#     SVM_SYNC_CCD_READ_37_50 = ("HeaterSync_ccdRead37_50.arb",)  # SVM/TCS sync signal (image cycle time: 37.50s)
#     SVM_SYNC_CCD_READ_43_75 = ("HeaterSync_ccdRead43_75.arb",)  # SVM/TCS sync signal (image cycle time: 43.75s)
#     SVM_SYNC_CCD_READ_50 = "HeaterSync_ccdRead50.arb"  # SVM/TCS sync signal (image cycle time: 50s)
#
#     SVM_SYNC_F_CAM = "HeaterSync_FccdRead.arb"  # SVM/TCS sync signal (fixed image cycle time)
#     SVM_SYNC_F_CAM_MIN_POINTS = "Heater_FccdRead_min_points.arb"  # Taken from PTO-EST-SC-TN-1563 (v1.0)
#
#
# class ArbData:
#     """This code is based on the code of ARB_DATA.py, developed by EVOLEO."""
#
#     def __init__(self):
#         """Initialisation of ARB data that can be sent to or be received from the AWG."""
#
#         self.array = []
#         self.filename = None
#
#     # def init_from_file(self, filename: ArbDataFile):
#     #     """Read the ARB data from the given file.
#     #
#     #     This consists of the following steps:
#     #
#     #         - Look for the file with the given name in the /arbdata folder;
#     #         - Read the data from this file;
#     #         - Convert the data to an array.
#     #
#     #     Args:
#     #         - filename: Filename as in the /arbdata folder.
#     #     """
#     #
#     #     self.filename = get_resource(f":/aeudata/{filename}")
#     #
#     #     self.parse_from_file()
#
#     def init_from_bytestring(self, bytestring: bytes):
#         """Read the ARB data from the given bytestring.
#
#         Args:
#             - bytestring: Bytestring.
#         """
#
#         bytestring = bytestring.decode(encoding="latin1", errors="ignore")
#
#         len_num_bytes = int(bytestring[1])
#         data = bytestring[2 + len_num_bytes :]
#
#         self.array_from_bytes(data)
#
#     def parse_from_file(self):
#         """Construct an array from the content of the file.
#
#         The first line of the file indicates what the format is (currently, all files start with "HEX").
#         """
#
#         with open(self.filename, "r") as arb_data_file:
#             # Check what the data format is
#
#             header = arb_data_file.readline()
#             args = header.strip().split("\t")
#             data_format = args[0]
#
#             # Data itself (which should be turned into an array)
#
#             data = arb_data_file.read()
#
#             if data_format == "HEX":
#                 self.array_from_hex_string(data)
#
#             else:
#                 from egse.arbitrary_wave_generator.aim_tti.tgf4000 import Tgf4000Error
#                 raise Tgf4000Error("The first line in the ARB data file should be: HEX")
#
#     def array_from_hex_string(self, hex_string):
#         """Build an array from the given hex string.
#
#         Examples:
#             - hex_string: "0001 0002 0003" -> array: [1, 2, 3]
#             - hex_string: "7FFF FFFF 0000 0001 8001" -> array: [32767, -1, 0, 1, -32767]
#             - hex_string: "000100020003" -> array: [1, 2, 3]
#
#         Args:
#             - hex_string: Hex string to create an array from.
#         """
#
#         self.array = []
#
#         # Strip off the blanks
#
#         hex_string = hex_string.replace(" ", "").strip()
#
#         # Each number is represented by 4 characters
#
#         num_hex_numbers = len(hex_string) // 4
#
#         # Loop over all hex numbers in the input string
#
#         for index in range(num_hex_numbers):
#             index *= 4
#             hex_number = hex_string[index : index + 4]
#
#             # print(f"Hex number: {hex_number} -> {int(np.int16(int(hex_number, 16)))}")
#
#             # Folding
#             # E.g. FFFF is folded to -1 (int returns 65535)
#
#             self.array.append(int(np.int16(int(hex_number, 16))))
#
#     @property
#     def string(self):
#         """Convert the content of the file to the ARB data as it should be sent to the AWG.
#
#         The returned string is a concatenation of:
#
#             - The #-symbol;
#             - The number of ASCII characters that will define the length of the binary data [bytes];
#             - The length of the binary data [bytes];
#             - The binary data itself.
#
#         Returns: Content of the file as the ARB data as it should be sent to the AWG.
#         """
#
#         # Convert the array to a byte string ("latin1": https://stackoverflow.com/questions/42795042)
#
#         byte_array = self.array_as_bytes().decode(encoding="latin1", errors="ignore")
#
#         # Number of ASCII characters that define the length of the binary data [bytes]
#
#         num_bytes = len(byte_array)
#
#         # Length of the binary data [bytes]
#
#         str_num_bytes = str(num_bytes)
#
#         # Binary data itself
#
#         len_num_bytes = len(str_num_bytes)
#
#         header = f"#{len_num_bytes:1d}{str_num_bytes}"
#
#         # return header.encode() + byte_array
#         return rf"#{len_num_bytes:1d}{str_num_bytes}{byte_array}"
#
#         # return rf"#{len_num_bytes:1d}{str_num_bytes}{byte_array.hex()}"
#
#     def array_as_bytes(self):
#         """Return the current array as a byte array.
#
#         Returns: Current array as a byte array
#         """
#
#         byte_string = bytes()
#
#         for number in self.array:
#             byte_string += number.to_bytes(length=2, byteorder="big", signed=True)
#
#         return byte_string
#
#     def array_from_bytes(self, data: str):
#         """Parse array value from the given string.
#
#         Args:
#             - data: Data string.
#         """
#
#         num_array = []
#
#         if isinstance(data, str):
#             bytestring = data.encode(encoding="latin1")
#
#         num_numbers = len(bytestring) // 2
#
#         for index in range(num_numbers):
#             index *= 2
#             number = bytestring[index : index + 2]
#
#             num_array.append(int.from_bytes(number, byteorder="big", signed=True))
#
#         self.array = num_array


class ArbData:
    """This code is based on the code of ARB_DATA.py, developed by EVOLEO."""

    def __init__(self):
        """Initialisation of ARB data that can be sent to or be received from the Aim-TTi TGF4000."""

        self.array = []
        self.filename = None

    def init_from_file(self, filename: str) -> None:
        """Read the ARB data from the given file.

        This consists of the following steps:

            - Look for the file with the given name in the /arbdata folder;
            - Read the data from this file;
            - Convert the data to an array.

        Args:
            filename (str): Filename.
        """

        self.filename = filename
        self.parse_from_file()

    def init_from_bytestring(self, bytestring: bytes) -> None:
        """Reads the ARB data from the given bytestring.

        Args:
            bytestring (bytes): Bytestring.
        """

        bytestring = bytestring.decode(encoding="latin1", errors="ignore")

        len_num_bytes = int(bytestring[1])
        data = bytestring[2 + len_num_bytes :]

        self.array_from_bytes(data)

    def parse_from_file(self) -> None:
        """Constructs an array from the content of the file.

        The first line of the file indicates what the format is (currently, all files start with "HEX").
        """

        with open(self.filename, "r") as arb_data_file:
            # Check what the data format is

            header = arb_data_file.readline()
            args = header.strip().split("\t")
            data_format = args[0]

            # Data itself (which should be turned into an array)

            data = arb_data_file.read()

            if data_format == "HEX":
                self.array_from_hex_string(data)
            else:
                from egse.arbitrary_wave_generator.aim_tti.tgf4000 import Tgf4000Error

                raise Tgf4000Error("The first line in the ARB data file should be: HEX")

    def array_from_hex_string(self, hex_string) -> None:
        """Builds an array from the given hex string.

        Examples:
            - hex_string: "0001 0002 0003" -> array: [1, 2, 3]
            - hex_string: "7FFF FFFF 0000 0001 8001" -> array: [32767, -1, 0, 1, -32767]
            - hex_string: "000100020003" -> array: [1, 2, 3]

        Args:
            hex_string (str): Hex string to create an array from.
        """

        self.array = []

        # Strip off the blanks

        hex_string = hex_string.replace(" ", "").strip()

        # Each number is represented by 4 characters

        num_hex_numbers = len(hex_string) // 4

        # Loop over all hex numbers in the input string

        for index in range(num_hex_numbers):
            index *= 4
            hex_number = hex_string[index : index + 4]

            # print(f"Hex number: {hex_number} -> {int(np.int16(int(hex_number, 16)))}")

            # Folding
            # E.g. FFFF is folded to -1 (int returns 65535)

            self.array.append(int(np.int16(int(hex_number, 16))))

    @property
    def string(self):
        """Converts the content of the file to the ARB data as it should be sent to the Aim-TTi TGF4000.

        The returned string is a concatenation of:

            - The #-symbol,
            - The number of ASCII characters that will define the length of the binary data [bytes],
            - The length of the binary data [bytes],
            - The binary data itself.

        Returns:
            Content of the file as the ARB data as it should be sent to the Aim-TTi TGF4000.
        """

        # Convert the array to a byte string ("latin1": https://stackoverflow.com/questions/42795042)

        byte_array = self.array_as_bytes().decode(encoding="latin1", errors="ignore")

        # Number of ASCII characters that define the length of the binary data [bytes]

        num_bytes = len(byte_array)

        # Length of the binary data [bytes]

        str_num_bytes = str(num_bytes)

        # Binary data itself

        len_num_bytes = len(str_num_bytes)

        header = f"#{len_num_bytes:1d}{str_num_bytes}"

        # return header.encode() + byte_array
        return rf"#{len_num_bytes:1d}{str_num_bytes}{byte_array}"

        # return rf"#{len_num_bytes:1d}{str_num_bytes}{byte_array.hex()}"

    def array_as_bytes(self) -> bytes:
        """Returns the current array as a byte array.

        Returns:
            Current array as a byte array.
        """

        byte_string = bytes()

        for number in self.array:
            byte_string += number.to_bytes(length=2, byteorder="big", signed=True)

        return byte_string

    def array_from_bytes(self, data: str) -> None:
        """Parses array value from the given string.

        Args:
            data (str): Data string.
        """

        num_array = []

        if isinstance(data, str):
            bytestring = data.encode(encoding="latin1")

        num_numbers = len(bytestring) // 2

        for index in range(num_numbers):
            index *= 2
            number = bytestring[index : index + 2]

            num_array.append(int.from_bytes(number, byteorder="big", signed=True))

        self.array = num_array


class WaveformShape(str, Enum):
    """Enumeration of the waveform shapes."""

    SINE = "SINE"  # Sinusoidal waveform
    SQUARE = "SQUARE"  # Square waveform
    RAMP = "RAMP"
    RAMPUP = RAMP_UP = "RAMPUP"
    RAMPDOWN = RAMP_DOWN = "RAMPDOWN"
    TRIANGULAR = TRIANG = TRIANGLE = "TRIANG"  # Triangular waveform
    PULSE = "PULSE"
    NOISE = "NOISE"  # Gaussian white noise waveform
    PRBSPN7 = "PRBSPN7"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 7 bits)
    PRBSPN9 = "PRBSPN9"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 9 bits)
    PRBSPN11 = "PRBSPN11"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 11 bits)
    PRBSPN15 = "PRBSPN15"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 15 bits)
    PRBSPN20 = "PRBSPN20"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 20 bits)
    PRBSPN23 = "PRBSPN23"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 23 bits)
    PRBSPN29 = "PRBSPN29"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 29 bits)
    PRBSPN31 = "PRBSPN31"  # Pseudo-Random Binary Sequence (PRBS) waveform (LFSR length: 31 bits)
    ARBITRARY = ARB = "ARB"  # Arbitrary waveform


class AmplitudeRange(str, Enum):
    """Enumeration of the amplitude ranges."""

    AUTO = "AUTO"
    HOLD = "HOLD"


class OutputWaveformType(str, Enum):
    """Enumeration of the output waveform types."""

    DC = "DC"
    SINC = "SINC"
    HAVERSINE = "HAVERSINE"
    CARDIAC = "CARDIAC"
    EXP_RISE = EXPRISE = "EXPRISE"  # Exponential rise
    LOG_RISE = LOGRISE = "LOGRISE"  # Logarithmic rise
    EXP_FALL = EXPFALL = "EXPFALL"  # Exponential fall
    LOG_FALL = LOGFALL = "LOGFALL"  # Logarithmic fall
    GAUSSIAN = GAUSS = "GAUSSIAN"
    LORENTZ = "LORENTZ"
    D_LORENTZ = DLORENTZ = "DLORENTZ"
    TRIANGULAR = TRIANG = TRIANGLE = "TRIANG"
    ARB1 = "ARB1"  # Arbitrary waveform specified and stored by the user in memory under "ARB1"
    ARB2 = "ARB2"  # Arbitrary waveform specified and stored by the user in memory under "ARB2"
    ARB3 = "ARB3"  # Arbitrary waveform specified and stored by the user in memory under "ARB3"
    ARB4 = "ARB4"  # Arbitrary waveform specified and stored by the user in memory under "ARB4"


class Modulation(str, Enum):
    """Enumeration of the modulation types."""

    OFF = "OFF"
    AM = "AM"  # Amplitude Modulation
    AMSC = AM_SC = "AMSC"  # Amplitude Modulation - Suppressed Carrier
    FM = "FM"  # Frequency Modulation
    PM = "PM"  # Phase Modulation
    ASK = "ASK"  # Amplitude Shift Keying
    FSK = "FSK"  # Frequency Shift Keying
    SUM = "SUM"  # Carrier + Modulating signal
    BPSK = "BPSK"  # Binary Phase Shift Keying
    PWM = "PWM"  # Pulse Width Modulation


class Output(str, Enum):
    """Enumeration of outputs."""

    ON = "ON"
    OFF = "OFF"
    NORMAL = "NORMAL"
    INVERT = "INVERT"


class SyncOutput(str, Enum):
    """Enumeration of synchronisation output states."""

    OFF = "OFF"
    ON = "ON"


class SyncType(str, Enum):
    """Enumeration of synchronisation types."""

    AUTO = "AUTO"
    NORMAL = "NORMAL"
    CARRIER = "CARRIER"
    TRIGGER = "TRIGGER"
    OFF = "OFF"


class Channel2Config(str, Enum):
    """Enumeration of configurations for channel 2."""

    MAINOUT = MAIN_OUT = "MAINOUT"
    SYNCOUT = SYNC_OUT = "SYNCOUT"


class WaveformSource(str, Enum):
    """Enumeration of waveform sources."""

    INT = INTERNAL = "INT"
    EXT = EXTERNAL = "EXT"


class Slope(str, Enum):
    """Enumeration of slopes."""

    POS = POSITIVE = "POS"
    NEG = NEGATIVE = "NEG"


class SweepType(str, Enum):
    """Enumeration of sweep types."""

    LINUP = LIN_UP = LINEAR_UP = "LINUP"
    LINDN = LINDOWN = LIN_DOWN = LINEARDN = LINEARDOWN = LINEAR_DOWN = "LINDN"
    LOGUP = LOG_UP = "LOGUP"
    LOGDN = LOGDOWN = LOG_DN = LOG_DOWN = "LOGDN"


class SweepMode(str, Enum):
    """Enumeration of sweep modes."""

    CONT = CONTINUOUS = "CONT"
    TRIG = TRIGGER = "TRIG"


class TriggerSource(str, Enum):
    """Enumeration of trigger sources."""

    INT = INTERNAL = "INT"
    EXT = EXTERNAL = "EXT"
    MAN = MANUAL = "MAN"


class Sweep(str, Enum):
    """Enumeration of synchronisation sweep states."""

    OFF = "OFF"
    ON = "ON"


class Burst(str, Enum):
    """Enumeration of bursts"""

    OFF = "OFF"
    NCYC = "NCYC"
    GATED = "GATED"
    INFINITE = "INFINITE"


class ClockSource(str, Enum):
    """Enumeration of clock sources."""

    INT = INTERNAL = "INT"
    EXT = EXTERNAL = "EXT"


class ChannelMode(str, Enum):
    """Enumeration of manual trigger operations."""

    SINGLE = "SINGLE"
    DUAL = "DUAL"


class AmplitudeCoupling(str, Enum):
    """Enumeration of amplitude coupling."""

    ON = "ON"
    OFF = "OFF"


class OutputCoupling(str, Enum):
    """Enumeration of output coupling."""

    ON = "ON"
    OFF = "OFF"


class FrequencyCoupling(str, Enum):
    """Enumeration of frequency coupling."""

    ON = "ON"
    OFF = "OFF"


class ChannelTracking(str, Enum):
    """Enumeration of channel tracking."""

    OFF = "OFF"
    EQUAL = "EQUAL"
    INVERT = "INVERT"


class BeepMode(str, Enum):
    """Enumeration of beep modes."""

    ON = "ON"
    OFF = "OFF"
    WARN = WARNING = "WARN"
    ERROR = "ERROR"


class NetworkConfig(str, Enum):
    """Enumeration of network configurations."""

    DHCP = "DHCP"
    AUTO = "AUTO"
    STATIC = "STATIC"


class FilterShape(str, Enum):
    """Enumeration of arbitrary filter shapes"""

    NORMAL = "NORMAL"
    STEP = "STEP"


class CounterSource(str, Enum):
    """Enumeration of the input counter source."""

    AC = "AC"  # AC-coupled (Alternating Current)
    DC = "DC"  # DC-coupled (Direct Current)


class CounterType(str, Enum):
    """Enumeration of the counter type."""

    FREQUENCY = "FREQUENCY"  # Measure the frequency of the signal
    PERIOD = "PERIOD"  # Measure the period of the signal
    WIDTH = POS_WIDTH = "WIDTH"  # Measure the positive width of the signal
    NWIDTH = NEG_WIDTH = "NWIDTH"  # Measure the negative width of the signal
    DUTY = "DUTY"  # Measure the duty cycle of the signal



"""
Part of the housekeeping acquired during the Ariel TA test campaign at CSL will be stored directly into the MySQL
facility database.  This module enables watching specific tables in that database for new entries.  The goal is to
store those into the dedicated CSV files for the TA-EGSE framework.
"""

import threading
from datetime import datetime

from egse.confman.confman_cs import load_setup
from egse.hk import read_conversion_dict, convert_hk_names
from egse.log import egse_logger
from egse.system import str_to_datetime, format_datetime
from egse.metrics import get_metrics_repo
from egse.settings import Settings, get_site_id
from pymysqlreplication import BinLogStreamReader
from pymysqlreplication.row_event import WriteRowsEvent
from urllib3.exceptions import NewConnectionError
import os

from egse.storage import StorageProxy

LOGGER = egse_logger
SITE_ID = get_site_id()
FACILITY_DB_SETTINGS = Settings.load("Facility DB")

ID_COLUMN_NAME = "measure_id"
TIMESTAMP_COLUMN_NAME = "measure_timestamp"
VALUE_COLUMN_NAME = "measure_value"


class DatabaseTableWatcher:
    def __init__(self, table_name: str, origin: str, server_id: int):
        """Initialisation of a watcher for a specific table in the facility database.

        The watcher is a daemon thread the watches the specified table in the facility database for new entries.  If
        a new entry is encountered, it will be sent to the Storage Manager, which will store it in the HK file for the
        given storage mnemonic.

        Args:
            table_name (str): Name of the table in the facility database.
            origin (str): Storage mnemonic for the data in the TA-EGSE framework.
            server_id (int): Unique identifier for the MySQL binlog stream reader.
        """

        self.table_name = table_name
        self.origin = origin
        self.server_id = server_id

        # noinspection PyBroadException
        try:
            self.hk_conversion_dict = read_conversion_dict(self.origin, use_site=False, setup=load_setup())
        except:
            self.hk_conversion_dict = None

        # Make a thread and let it start watching the specified table in the facility database

        self.watch_thread = threading.Thread(target=self.watch_db_table)
        self.watch_thread.daemon = True
        self.keep_watching = True

        # Metrics client

        token = os.getenv("INFLUXDB3_AUTH_TOKEN")
        project = os.getenv("PROJECT")

        if project and token:
            self.metrics_client = get_metrics_repo(
                "influxdb", {"host": "http://localhost:8181", "database": project, "token": token}
            )
            self.metrics_client.connect()
        else:
            self.metrics_client = None
            LOGGER.warning(
                "INFLUXDB3_AUTH_TOKEN and/or PROJECT environment variable is not set.  Metrics will not be propagated "
                "to InfluxDB."
            )

    def start_watching_db_table(self):
        """Starts the thread that checks for new entries in the specified table in the facility database."""

        self.keep_watching = True
        self.watch_thread.start()

    def stop_watching_db_table(self):
        """Stops the thread that checks for new entries in the specified table in the facility database."""

        self.keep_watching = False
        self.watch_thread.join()

    def watch_db_table(self):
        """Lets the thread watch for new entries in the specified table in the facility database.

        If a new entry is encountered, it will be sent to the Storage Manager, which will store it in the HK file for
        the specified storage mnemonic.
        """

        # Connect to the (MySQL) facility database -> Start watching the specified table

        mysql_settings = {
            "host": FACILITY_DB_SETTINGS.HOST,
            "port": FACILITY_DB_SETTINGS.PORT,
            "user": FACILITY_DB_SETTINGS.USER,
            "passwd": FACILITY_DB_SETTINGS.PASSWORD,
        }

        stream: BinLogStreamReader = BinLogStreamReader(
            connection_settings=mysql_settings,
            server_id=self.server_id,  # Unique identifier per Python client watching the MySQL facility database
            blocking=True,
            only_events=[WriteRowsEvent],  # Watch for new row entries
            only_tables=[self.table_name],  # Watch for changes in the specified table
        )

        while self.keep_watching:
            for bin_log_event in stream:
                for row in bin_log_event.rows:
                    values = row["values"]  # Dictionary with the column names (from the facility database) as keys
                    hk = self.translate_parameter_names(values)  # Convert to TA-EGSE-consistent names
                    self.store_housekeeping_information(hk)
                    self.propagate_metrics(hk)

        stream.close()

    def translate_parameter_names(self, hk: dict):
        """Converts the parameter names from the facility database to TA-EGSE-consistent names.

        Args:
            hk (dict): Dictionary with the column names (from the facility database) as keys.

        Returns:
            Dictionary with the TA-EGSE-consistent names as keys.
        """

        # Timestamp

        # noinspection PyUnresolvedReferences
        hk["timestamp"] = format_datetime(
            datetime.datetime.fromtimestamp(hk[TIMESTAMP_COLUMN_NAME], datetime.UTC)
        )  # Unix time -> datetime [UTC]
        del hk[TIMESTAMP_COLUMN_NAME]

        # Delete identifier of the entry

        del hk[ID_COLUMN_NAME]

        # Parameter value

        hk[self.table_name] = hk[VALUE_COLUMN_NAME]
        del hk[VALUE_COLUMN_NAME]

        if self.hk_conversion_dict:
            return convert_hk_names(hk, self.hk_conversion_dict)
        else:
            return hk

    def store_housekeeping_information(self, hk: dict):
        """Sends the given housekeeping information to the Storage Manager.

        The housekeeping is passed as a dictionary, with the parameter names as keys.  There's also an entry for the
        timestamp, which represents the date/time at which the value was received.

        Args:
            hk (dict): Housekeeping that was extracted from the facility database, after converting the parameter names
                       to TA-EGSE-consistent names.
        """

        try:
            with StorageProxy() as storage:
                response = storage.save({"origin": self.origin, "data": hk})
                if not response.successful:
                    LOGGER.warning(
                        f"Couldn't save facility data to the Storage manager for {self.origin}, cause: {response}"
                    )
        except ConnectionError as exc:
            LOGGER.warning(
                f"Couldn't connect to the Storage Manager to store facility housekeeping for {self.origin}: {exc}"
            )
            raise

    def propagate_metrics(self, hk: dict):
        """Propagates the given housekeeping information to the metrics database.

        The housekeeping is passed as a dictionary, with the parameter names as keys.  There's also an entry for the
        timestamp, which represents the date/time at which the value was received.  In case only the timestamp is
        present in the dictionary, nothing will be written to the metrics database.

        Args:
            hk (dict): Housekeeping that was extracted from the facility database, after converting the parameter names
                       to TA-EGSE-consistent names.
        """

        if not [x for x in hk if x != "timestamp"]:
            LOGGER.debug(f"no metrics defined for {self.origin}")
            return

        try:
            if self.metrics_client:
                point = {
                    "measurement": self.origin.lower(),
                    "tags": {"site_id": SITE_ID, "origin": self.origin},
                    "fields": {hk_name.lower(): hk[hk_name] for hk_name in hk if hk_name != "timestamp"},
                    "time": str_to_datetime(hk["timestamp"]),
                }
                self.metrics_client.write(point)
            else:
                LOGGER.warning(
                    f"Could not write {self.origin} metrics to the time series database (self.metrics_client is None)."
                )
        except NewConnectionError:
            LOGGER.warning(
                f"No connection to the time series database could be established to propagate {self.origin} metrics.  "
                f"Check whether this service is (still) running."
            )



import logging
from pathlib import Path
from typing import Optional

import h5py
import natsort
import numpy as np
from egse.spw import DataDataPacket
from egse.spw import HousekeepingPacket
from egse.spw import OverscanDataPacket
from egse.spw import TimecodePacket
from egse.persistence import (
    PersistenceLayer,
)  # -> circular dependency because this import loads this module as a plugin

LOGGER = logging.getLogger(__name__)


class HDF5(PersistenceLayer):
    extension = "hdf5"

    def __init__(self, filename, prep: dict = None):
        """
        The `prep` argument needs at least the following mandatory key:value pairs:

          * mode: the mode used for opening the file [default is 'r']

        """
        # LOGGER.debug(f"{h5py.version.hdf5_version=}")
        self._filepath = Path(filename).expanduser()
        self._mode = prep.get("mode") or "r"
        self._h5file: Optional[h5py.File] = None

    def __enter__(self):
        self.open(mode=self._mode)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self, mode=None):
        self._mode = mode or self._mode
        LOGGER.debug(f"Opening file {self._filepath} in mode '{self._mode}'")
        self._h5file = h5py.File(self._filepath, mode=self._mode)

        # File "h5py/h5f.pyx", line 554, in h5py.h5f.FileID.start_swmr_write
        # RuntimeError: Unable to start swmr writing (file superblock version - should be at least 3)
        # self._h5file.swmr_mode = True

    def close(self):
        self._h5file.close()

    def exists(self):
        return self._filepath.exists()

    def create(self, data):
        """
        Store the given data in the HDF5 file. The data argument shall be a dictionary where the
        keys represent the group where the data shall be saved, and the value is the data to be
        saved. When the key ends with ":ATTRS", then the value is a list of attributes to that
        group. Values can be of different type and are processed if needed.

        An example data argument:

            {
                "/10/timecode": tc_packet,
                "/10/timecode:ATTRS": [("timestamp", timestamp)],
                "/10/command/": f"{command.__name__}, {args=}",
                "/10/register/": self.register_map.get_memory_map_as_ndarray()
            }

        The example saves a Timecode packet in the group "/10/timecode" and attaches a timestamp
        as an attribute called "timestamp" to the same group. It then adds a command string in
        the "/10/command" group and finally adds a register memory map (an np.ndarray) in the group
        "/10/register".

        Args:
            data (dict): a dictionary containing the data that needs to be saved.

        Returns:
            None.
        """
        for key, value in data.items():
            if key.endswith(":ATTRS"):
                a_key = key.split(":")[0]
                for k, v in value:
                    self._h5file[a_key].attrs[k] = v
            if isinstance(value, TimecodePacket):
                self._h5file[key] = value.timecode
            if isinstance(value, HousekeepingPacket):
                self._h5file[key] = value.packet_as_ndarray
            if isinstance(value, HousekeepingData):
                self._h5file[key] = value.data_as_ndarray
            if isinstance(value, DataDataPacket):
                self._h5file[key] = value.packet_as_ndarray
            if isinstance(value, OverscanDataPacket):
                self._h5file[key] = value.packet_as_ndarray
            if isinstance(value, (str, bytearray, np.ndarray)):
                # if we save a command, put it into a 'commands' group.
                # This is a special case that is the result of issue #1461

                if "command" in key:
                    idx = key.split("/")[1]
                    if idx in self._h5file and "commands" in self._h5file[idx]:
                        last_idx = int(sorted(self._h5file[f"/{idx}/commands"].keys())[-1])
                        key = f"/{idx}/commands/{last_idx + 1}"
                    else:
                        key = f"/{idx}/commands/0"

                self._h5file[key] = value

    def read(self, select=None):
        """
        Read information or data from the HDF5 file.

        The `select` argument can contain the following information:

        * the string 'number_of_groups': request to determine the number of top groups in
          the HDF5 file.
        * the string 'last_top_group': request the name/key of the last item in the top group.
                The last item is the last element of the list of keys, sorted with natural order.

        Args:
            select (str or dict): specify which information should be read

        Returns:
            When 'number_of_groups', return an integer, when 'last_top_group' return a string.
        """
        if select == "number_of_groups":
            return len(self._h5file.keys())
        if select == "last_top_group":
            keys = self._h5file.keys()

            LOGGER.debug(f"{self._h5file.filename}: {keys=}")

            return 0 if len(keys) == 0 else natsort.natsorted(keys)[-1]

            # This following lines is a longer version of the previous two lines, keep them for
            # debugging because I had problems and not yet sure what is the cause...

            # sorted_keys = natsort.natsorted(keys)
            # LOGGER.debug(f"{self._h5file.filename}: {sorted_keys=}")
            # key = sorted_keys[-1]
            # LOGGER.debug(f"{key=}")
            # return key

    def update(self, idx, data):
        pass

    def delete(self, idx):
        pass

    def get_filepath(self):
        return self._filepath



import random
import threading
import time
from enum import Enum
from typing import Any

from egse.env import bool_env
from egse.log import logging
from egse.system import type_name
from egse.zmq_ser import connect_address

logger = logging.getLogger("egse.connect")

# random.seed(time.monotonic())  # uncomment for testing only, main application should set a seed.

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")


def get_endpoint(
    service_type: str | None = None,
    protocol: str = "tcp",
    hostname: str = "localhost",
    port: int = 0,
):
    """
    Returns the endpoint for a service, either from the registry or by constructing
    it from protocol, hostname and port.

    If port is 0 (the default), attempt to retrieve the endpoint from the service registry.

    Args:
        service_type: The service type to look up in the registry.
        protocol: Protocol to use if constructing the endpoint, defaults to tcp.
        hostname: Hostname to use if constructing the endpoint, defaults to localhost.
        port: Port to use if constructing the endpoint, defaults to 0.

    Returns:
        The endpoint string.

    Raises:
        RuntimeError: If no endpoint can be determined.
    """
    endpoint = None
    from egse.registry.client import RegistryClient

    if port == 0:
        with RegistryClient() as reg:
            endpoint = reg.get_endpoint(service_type)
        if endpoint:
            if VERBOSE_DEBUG:
                logger.debug(f"Endpoint for '{service_type}' found in registry: {endpoint}")
        else:
            logger.warning(f"No endpoint for '{service_type}' found in registry.")

    if not endpoint:
        if port == 0:
            raise RuntimeError(f"No service registered as '{service_type}' and no port provided.")
        endpoint = connect_address(protocol, hostname, port)
        if VERBOSE_DEBUG:
            logger.debug(f"Endpoint constructed from protocol/hostname/port: {endpoint}")

    return endpoint


class ConnectionState(Enum):
    DISCONNECTED = "disconnected"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    CIRCUIT_OPEN = "circuit_open"  # Temporarily stopped trying


class BackoffStrategy(Enum):
    """
    Specifies the strategy for increasing the delay between retry attempts
    in backoff algorithms to reduce load and avoid overwhelming services.

    Strategies:
        EXPONENTIAL:
            The delay doubles with each retry attempt (e.g., 1s, 2s, 4s, 8s).
            This is the most widely used approach because it quickly reduces load on struggling systems.
        LINEAR:
            The delay increases by a fixed amount each time (e.g., 1s, 2s, 3s, 4s).
            This provides a more gradual reduction in request rate.
        FIXED:
            Uses the same delay between all retry attempts.
            Simple but less adaptive to system conditions.

    References:
        - AWS Architecture Blog: Exponential Backoff And Jitter
    """

    EXPONENTIAL = "exponential"
    """The delay doubles with each retry attempt (e.g., 1s, 2s, 4s, 8s). 
    This is the most widely used approach because it quickly reduces load on struggling systems."""
    LINEAR = "linear"
    """The delay increases by a fixed amount each time (e.g., 1s, 2s, 3s, 4s). 
    This provides a more gradual reduction in request rate."""
    FIXED = "fixed"
    """Uses the same delay between all retry attempts. Simple but less adaptive to system conditions."""


class JitterStrategy(Enum):
    """
    Specifies the strategy for applying jitter (randomization) to retry intervals
    in backoff algorithms to avoid synchronized retries and reduce load spikes.

    Strategies:
        NONE:
            No jitter is applied. The retry interval is deterministic.
        FULL:
            Applies full jitter by selecting a random value uniformly between 0 and the calculated interval.
            This maximizes randomness but can result in very short delays.
        EQUAL:
            Applies "equal jitter" as described in the AWS Architecture Blog.
            The interval is randomized within [interval/2, interval], ensuring a minimum delay of half the interval.
            Note: This is not the same as "a jitter of 50% around interval" (which would be [0.5 * interval, 1.5 * interval]).
        PERCENT_10:
            Applies a jitter of ±10% around the base interval, resulting in a random interval within [0.9 * interval, 1.1 * interval].

    References:
        - AWS Architecture Blog: Exponential Backoff And Jitter
    """

    NONE = "none"
    """No jitter is applied to the backoff."""
    FULL = "full"
    """Maximum distribution but can be too random with very short intervals."""
    EQUAL = "equal"
    """Best balance, maintains backoff properties while preventing synchronization."""
    PERCENT_10 = "10%"
    """Add a jitter of 10% around the base interval."""


def calculate_retry_interval(
    attempt_number,
    base_interval,
    max_interval,
    backoff_strategy: BackoffStrategy = BackoffStrategy.EXPONENTIAL,
    jitter_strategy: JitterStrategy = JitterStrategy.EQUAL,
):
    """
    Calculates the next retry interval based on the given backoff and jitter strategies.

    Args:
        attempt_number (int): The current retry attempt (starting from 0).
        base_interval (float): The initial interval in seconds.
        max_interval (float): The maximum allowed interval in seconds.
        backoff_strategy (BackoffStrategy): Strategy for increasing the delay (exponential, linear, or fixed).
        jitter_strategy (JitterStrategy): Strategy for randomizing the delay to avoid synchronization.

    Returns:
        float: The computed retry interval in seconds.

    Notes:
        - See the docstrings for BackoffStrategy and JitterStrategy for details on each strategy.
        - Based on best practices from the AWS Architecture Blog: Exponential Backoff And Jitter.
    """

    if backoff_strategy == BackoffStrategy.EXPONENTIAL:
        interval = min(base_interval * (2**attempt_number), max_interval)
    elif backoff_strategy == BackoffStrategy.LINEAR:
        interval = min(base_interval + attempt_number, max_interval)
    else:
        interval = base_interval

    if jitter_strategy == JitterStrategy.NONE:
        return interval
    elif jitter_strategy == JitterStrategy.FULL:
        return random.uniform(0, interval)
    elif jitter_strategy == JitterStrategy.EQUAL:
        return interval / 2 + random.uniform(0, interval / 2)
    elif jitter_strategy == JitterStrategy.PERCENT_10:
        jitter_amount = interval * 0.1
        return interval + random.uniform(-jitter_amount, jitter_amount)

    return interval


class AsyncServiceConnector:
    """
    Asynchronous base class for robust service connection management with retry, backoff, and circuit breaker logic.

    This class is intended to be subclassed for managing persistent connections to external services
    (such as devices, databases, or remote APIs) that may be unreliable or temporarily unavailable.

    Features:
        - Automatic retry with configurable backoff and jitter strategies.
        - Circuit breaker to prevent repeated connection attempts after multiple failures.
        - Connection state tracking (disconnected, connecting, connected, circuit open).

    Usage:
        1. Subclass `AsyncServiceConnector` and override the `connect_to_service()` coroutine with your
           actual connection logic. Optionally, override `health_check()` for custom health verification.
        2. Store the actual connection object (e.g., socket, transport) as an instance attribute in your subclass.
        3. Use `attempt_connection()` to initiate connection attempts; it will handle retries and backoff automatically.
        4. Use `is_connected()` to check connection status.

    Example:
        class MyConnector(AsyncServiceConnector):
            async def connect_to_service(self):
                self.connection = await create_socket()
                return self.connection is not None

            def get_connection(self):
                return self.connection

    Note:
        The base class does not manage or expose the underlying connection object.
        Your subclass should provide a method or property to access it as needed.
    """

    def __init__(
        self,
        service_name: str,
        backoff_strategy: BackoffStrategy = BackoffStrategy.EXPONENTIAL,
        jitter_strategy: JitterStrategy = JitterStrategy.EQUAL,
    ):
        self.state = ConnectionState.DISCONNECTED
        self.last_attempt = 0
        self.base_interval = 1
        self.retry_interval = 1  # Start with 1 second
        self.max_retry_interval = 300  # Max 5 minutes
        self.failure_count = 0
        self.max_failures_before_circuit_break = 5
        self.circuit_break_duration = 60  # 1 minute
        self.circuit_opened_at = None
        self.backoff_strategy = backoff_strategy
        self.jitter_strategy = jitter_strategy

        self.service_name = service_name

    async def connect_to_service(self) -> bool:
        logger.warning(
            f"The connect_to_service() method is not implemented for {self.service_name}, connection will always fail."
        )
        return False

    async def disconnect_from_service(self) -> None:
        """
        Optional hook to cleanly disconnect / release resources for the service.
        Default implementation is a no-op. Subclasses should override to:
          - close async transports
          - cancel background tasks
          - set state to ConnectionState.DISCONNECTED
          - call device.disconnect()
        """
        logger.debug(f"{self.service_name}: default async disconnect_from_service(): no-op")
        self.state = ConnectionState.DISCONNECTED
        return

    async def health_check(self) -> bool:
        logger.warning(
            f"The health_check() method is not implemented for {self.service_name}, check will always return false."
        )
        return False

    def should_attempt_connection(self) -> bool:
        """Return True if we should attempt a new connection."""
        now = time.monotonic()

        # If circuit is open, check if we should close it
        if self.state == ConnectionState.CIRCUIT_OPEN:
            assert self.circuit_opened_at is not None
            circuit_break_open_since = now - self.circuit_opened_at
            logger.debug(f"{circuit_break_open_since=}")
            if circuit_break_open_since > self.circuit_break_duration:
                self.state = ConnectionState.DISCONNECTED
                self.failure_count = 0
                self.retry_interval = 1
                return True
            return False

        # Regular backoff logic
        return now - self.last_attempt >= self.retry_interval

    async def attempt_connection(self):
        """Try to connect to the service.

        This will execute the `connect_to_service()` that was overridden by the subclass.
        That function shall return True when the connection succeeded, False otherwise.
        """
        if self.state == ConnectionState.CONNECTED:
            # ensure the CONNECTED state is validated before skipping reconnection attempts
            # even is state is CONNECTED, the underlying connection could be stale or broken
            # or closed externally and unless you check the health here, you will never attempt
            # recovery.
            try:
                healthy = await self.health_check()
            except Exception as exc:
                logger.debug(f"health_check raised: {type_name(exc)} – {exc}")
                healthy = False

            if healthy:
                if VERBOSE_DEBUG:
                    logger.debug(f"{self.service_name} already connected and healthy")
                return

            logger.info(
                f"{self.service_name} marked CONNECTED but health_check failed — disconnecting and reconnecting"
            )
            self.state = ConnectionState.DISCONNECTED
            try:
                # ensure the state is updated by disconnect hook (disconnect_from_service should set DISCONNECTED)
                await self.disconnect_from_service()
            except Exception as exc:
                if VERBOSE_DEBUG:
                    logger.debug(f"Couldn't disconnect from {self.service_name}")

        if not self.should_attempt_connection():
            logger.debug("Not time yet to attempt new connection")
            return

        self.state = ConnectionState.CONNECTING
        self.last_attempt = time.monotonic()

        try:
            success = await self.connect_to_service()

            if success:
                self.state = ConnectionState.CONNECTED
                self.failure_count = 0
                self.retry_interval = 1  # Reset backoff
                logger.info(f"Successfully connected to service {self.service_name}")
            else:
                # warning should have been logged by the connect_to_service() callable.
                self.handle_connection_failure()

        except Exception as exc:
            logger.warning(f"Failed to connect to service {self.service_name}: {exc}")
            self.handle_connection_failure()

    def handle_connection_failure(self):
        self.failure_count += 1

        # Open circuit breaker if too many failures
        if self.failure_count >= self.max_failures_before_circuit_break:
            self.state = ConnectionState.CIRCUIT_OPEN
            self.circuit_opened_at = time.monotonic()
            logger.warning(
                f"Circuit breaker opened for service {self.service_name} after {self.failure_count} failures"
            )
        else:
            self.state = ConnectionState.DISCONNECTED
            self.retry_interval = calculate_retry_interval(
                self.failure_count,
                self.base_interval,
                self.max_retry_interval,
                self.backoff_strategy,
                self.jitter_strategy,
            )
            logger.debug(f"retry_interval={self.retry_interval}")

    def is_connected(self) -> bool:
        if VERBOSE_DEBUG:
            logger.debug(f"Checking if {self.service_name} is connected: {self.state.name}")
        return self.state == ConnectionState.CONNECTED

    def get_connection(self) -> Any:
        """
        Optional method to return the underlying connection object.
        Subclasses should override this method to return the actual connection
        (e.g., socket, transport) if needed.
        """
        logger.warning(f"The get_connection() method is not implemented for {self.service_name}, returning None.")
        return None


class ServiceConnector:
    """
    Synchronous base class for robust service connection management with retry, backoff, and circuit breaker logic.

    This class is intended to be subclassed for managing persistent connections to external services
    (such as devices, databases, or remote APIs) that may be unreliable or temporarily unavailable.

    Features:
        - Automatic retry with configurable backoff and jitter strategies.
        - Circuit breaker to prevent repeated connection attempts after multiple failures.
        - Connection state tracking (disconnected, connecting, connected, circuit open).
        - Thread-safe operation using a lock for all state changes.

    Usage:
        1. Subclass `ServiceConnector` and override the `connect_to_service()` method with your
           actual connection logic. Optionally, override `health_check()` for custom health verification.
        2. Store the actual connection object (e.g., socket, transport) as an instance attribute in your subclass.
        3. Use `attempt_connection()` to initiate connection attempts; it will handle retries and backoff automatically.
        4. Use `is_connected()` to check connection status.

    Example:
        class MyConnector(ServiceConnector):
            def connect_to_service(self):
                self.connection = create_socket()
                return self.connection is not None

            def get_connection(self):
                return self.connection

    Note:
        The base class does not manage or expose the underlying connection object.
        Your subclass should provide a method or property to access it as needed.
    """

    def __init__(
        self,
        service_name: str,
        backoff_strategy: BackoffStrategy = BackoffStrategy.EXPONENTIAL,
        jitter_strategy: JitterStrategy = JitterStrategy.EQUAL,
    ):
        self.state = ConnectionState.DISCONNECTED
        self.last_attempt = 0
        self.base_interval = 1
        self.retry_interval = 1
        self.max_retry_interval = 300
        self.failure_count = 0
        self.max_failures_before_circuit_break = 5
        self.circuit_break_duration = 60
        self.circuit_opened_at = None
        self.service_name = service_name
        self.backoff_strategy = backoff_strategy
        self.jitter_strategy = jitter_strategy

        self._lock = threading.RLock()

    def connect_to_service(self) -> bool:
        logger.warning(
            f"The connect_to_service() method is not implemented for {self.service_name}, connection will always fail."
        )
        return False

    def disconnect_from_service(self) -> None:
        """
        Optional hook to cleanly disconnect / release resources for the service. Default implementation is a no-op.
        Subclasses should override and must be careful about thread-safety; the base class holds _lock which can be
        used.
        """
        with self._lock:
            logger.debug(f"{self.service_name}: default disconnect_from_service(): no-op")
            self.state = ConnectionState.DISCONNECTED
        return

    def health_check(self) -> bool:
        logger.warning(
            f"The health_check() method is not implemented for {self.service_name}, check will always return false."
        )
        return False

    def should_attempt_connection(self) -> bool:
        now = time.monotonic()
        with self._lock:
            if self.state == ConnectionState.CIRCUIT_OPEN:
                assert self.circuit_opened_at is not None
                if now - self.circuit_opened_at > self.circuit_break_duration:
                    self.state = ConnectionState.DISCONNECTED
                    self.failure_count = 0
                    self.retry_interval = 1
                    return True
                return False
            return now - self.last_attempt >= self.retry_interval

    def attempt_connection(self):
        with self._lock:
            current_state = self.state

        if current_state == ConnectionState.CONNECTED:
            # ensure the CONNECTED state is validated before skipping reconnection attempts
            try:
                healthy = self.health_check()
            except Exception as exc:
                logger.debug(f"health_check raised: {type_name(exc)} – {exc}")
                healthy = False

            if healthy:
                logger.debug(f"{self.service_name} already connected and healthy")
                return

            logger.info(
                f"{self.service_name} marked CONNECTED but health_check failed — disconnecting and reconnecting"
            )
            self.state = ConnectionState.DISCONNECTED
            try:
                # ensure the state is updated by disconnect hook (disconnect_from_service should set DISCONNECTED)
                self.disconnect_from_service()
            except Exception as exc:
                if VERBOSE_DEBUG:
                    logger.debug(f"Couldn't disconnect from {self.service_name}: {type_name(exc)} – {exc}")

        with self._lock:
            if not self.should_attempt_connection():
                return
            self.state = ConnectionState.CONNECTING
            self.last_attempt = time.monotonic()

        try:
            success = self.connect_to_service()
            with self._lock:
                if success:
                    self.state = ConnectionState.CONNECTED
                    self.failure_count = 0
                    self.retry_interval = 1
                    logger.debug(f"Successfully connected to service {self.service_name}")
                else:
                    self.handle_connection_failure()
        except Exception as exc:
            logger.error(f"Failed to connect to service {self.service_name}: {exc}")
            with self._lock:
                self.handle_connection_failure()

    def handle_connection_failure(self):
        self.failure_count += 1
        if self.failure_count >= self.max_failures_before_circuit_break:
            self.state = ConnectionState.CIRCUIT_OPEN
            self.circuit_opened_at = time.monotonic()
            logger.warning(
                f"Circuit breaker opened for service {self.service_name} after {self.failure_count} failures"
            )
        else:
            self.state = ConnectionState.DISCONNECTED
            self.retry_interval = calculate_retry_interval(
                self.failure_count,
                self.base_interval,
                self.max_retry_interval,
                self.backoff_strategy,
                self.jitter_strategy,
            )
            logger.debug(f"retry_interval={self.retry_interval}")

    def is_connected(self) -> bool:
        with self._lock:
            return self.state == ConnectionState.CONNECTED



import asyncio
import sys

import rich


async def run_all_status(full: bool = False, suppress_errors: bool = True):
    tasks = [
        asyncio.create_task(status_rm_cs(suppress_errors)),
        asyncio.create_task(status_nh_cs(suppress_errors)),
        asyncio.create_task(status_log_cs(suppress_errors)),
        asyncio.create_task(status_sm_cs(full, suppress_errors)),
        asyncio.create_task(status_cm_cs(suppress_errors)),
        asyncio.create_task(status_pm_cs(suppress_errors)),
    ]

    await asyncio.gather(*tasks)


async def status_rm_cs(suppress_errors):
    proc = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "egse.registry.server",
        "status",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_nh_cs(suppress_errors):
    proc = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "egse.notifyhub.server",
        "status",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_log_cs(suppress_errors):
    proc = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "egse.logger.log_cs",
        "status",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_sm_cs(full: bool = False, suppress_errors: bool = True):
    cmd = [sys.executable, "-m", "egse.storage.storage_cs", "status"]
    if full:
        cmd.append("--full" if full else "")

    proc = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_cm_cs(suppress_errors):
    proc = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "egse.confman.confman_cs",
        "status",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")


async def status_pm_cs(suppress_errors):
    proc = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "egse.procman.procman_cs",
        "status",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    rich.print(stdout.decode().rstrip())
    if stderr and not suppress_errors:
        rich.print(f"[red]{stderr.decode()}[/]")



"""Control Server for the Aim-TTi TGF4000."""

import logging
import multiprocessing
from typing import Annotated

import rich
import sys
import typer
import zmq

from egse.arbitrary_wave_generator.aim_tti import PROTOCOL, CS_SETTINGS
from egse.control import is_control_server_active, ControlServer
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.settings import Settings
from egse.storage import store_housekeeping_information
from egse.zmq_ser import connect_address, get_port_number

logger = logging.getLogger("egse.arbitrary_wave_generator.aim_tti.tgf4000")
DEVICE_SETTINGS = Settings.load("Aim-TTi TGF4000")


def is_tgf4000_cs_active(device_id: str, timeout: float = 0.5) -> bool:
    """Checks whether the Aim-TTi TGF4000 Control Server is running.

    Args:
        device_id (str): Device identifier, as per (local) settings and setup.
        timeout (float): Timeout when waiting for a reply [s].

    Returns:
        True if the Aim-TTi TGF4000 Control Server is running and replied with the expected answer; False otherwise.
    """

    commanding_port = CS_SETTINGS[device_id].get(
        "COMMANDING_PORT", 0
    )  # Commanding port (as per settings or dynamically assigned)
    hostname = CS_SETTINGS[device_id].get("HOSTNAME", "localhost")  # Hostname

    if commanding_port != 0:
        protocol = PROTOCOL
        port = commanding_port

    else:
        with RegistryClient() as reg:
            service_type = CS_SETTINGS[device_id].get("SERVICE_TYPE", "tgf4000_cs")
            service = reg.discover_service(service_type)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]

            else:
                return False

    # noinspection PyUnboundLocalVariable
    endpoint = connect_address(protocol, hostname, port)

    return is_control_server_active(endpoint, timeout)


class Tgf4000ControlServer(ControlServer):
    def __init__(self, device_id: str, simulator: bool = False):
        """Initialisation of a new Aim-TTi TGF4000 Control Server.

        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
            simulator (bool): Indicates whether to operate in simulator mode.
        """
        self.cs_settings = CS_SETTINGS[device_id]
        super().__init__()

        self.device_id = device_id
        process_name = CS_SETTINGS[device_id].get("PROCESS_NAME", "tgf4000_cs")
        service_type = CS_SETTINGS[device_id].get("SERVICE_TYPE", "tgf4000_cs")

        multiprocessing.current_process().name = (
            process_name  # Name under which it is registered in the service registry
        )

        self.logger = logger
        self.service_name = process_name
        self.service_type = service_type

        from egse.arbitrary_wave_generator.aim_tti.tgf4000_protocol import Tgf4000Protocol

        self.device_protocol = Tgf4000Protocol(self, device_id, simulator=simulator)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type)

    def get_communication_protocol(self) -> str:
        """Returns the communication protocol used Aim-TTi TGF4000 Control Server.

        Returns:
            Communication protocol used by the Aim-TTi TGF4000 Control Server, as specified in the settings.
        """

        return PROTOCOL

    def get_commanding_port(self) -> int:
        """Returns the commanding port used by the Aim-TTi TGF4000 Control Server.

        Returns:
            Commanding port used by the Aim-TTi TGF4000 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_cmd_sock) or self.cs_settings.get("COMMANDING_PORT", 0)

    def get_service_port(self) -> int:
        """Returns the service port used by the Aim-TTi TGF4000 Control Server.

        Returns:
            Service port used by the Aim-TTi TGF4000 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_service_sock) or self.cs_settings.get("SERVICE_PORT", 0)

    def get_monitoring_port(self) -> int:
        """Returns the monitoring port used by the Aim-TTi TGF4000 Control Server.

        Returns:
            Monitoring port used by the Aim-TTi TGF4000 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_mon_sock) or self.cs_settings.get("MONITORING_PORT", 0)

    def get_storage_mnemonic(self) -> str:
        """Returns the storage mnemonic used by the Aim-TTi TGF4000 Control Server.

        Returns:
            Storage mnemonic used by the Aim-TTi TGF4000 Control Server, as specified in the settings.
        """

        return self.cs_settings.get("STORAGE_MNEMONIC", "TGF4000")

    def is_storage_manager_active(self):
        """Checks whether the Storage Manager is active."""

        from egse.storage import is_storage_manager_active

        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Sends housekeeping information of the Aim-TTi TGF4000 to the Storage Manager."""

        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        """Registers the Control Server to the Storage Manager."""

        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            },
        )

    def unregister_from_storage_manager(self):
        """Unregisters the Control Server from the Storage Manager."""

        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def after_serve(self):
        self.deregister_service()


app = typer.Typer()


@app.command()
def start(
    device_id: Annotated[
        str, typer.Argument(help="Identifies the hardware controller (as per local settings and setup)")
    ],
    simulator: Annotated[
        bool,
        typer.Option("--simulator", "--sim", help="start the Aim-TTi TGF4000 Control Server in simulator mode"),
    ] = False,
) -> int:
    """Starts the Aim-TTi TGF4000 Control Server with the given identifier."""

    # noinspection PyBroadException
    try:
        controller = Tgf4000ControlServer(device_id, simulator)
        controller.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exc:
        exit_code = exc.code if hasattr(exc, "code") else 0
        print(f"System Exit with code {exc.code}")
        sys.exit(exit_code)
    except Exception:
        logger.exception(f"Cannot start the Aim-TTi TGF4000 {device_id} Control Server")

    return 0


@app.command()
def stop(
    device_id: Annotated[
        str, typer.Argument(help="Identifies the hardware controller (as per local settings and setup)")
    ],
) -> None:
    """Sends a `quit_server` command to the Aim-TTi TGF4000 Control Server."""

    service_type = CS_SETTINGS[device_id].get("SERVICE_TYPE", "tgf4000_cs")

    with RegistryClient() as reg:
        service = reg.discover_service(service_type)

        if service:
            proxy = ServiceProxy(protocol="tcp", hostname=service["host"], port=service["metadata"]["service_port"])
            proxy.quit_server()
        else:
            try:
                with Tgf4000Proxy(device_id) as tgf4000_proxy:
                    with tgf4000_proxy.get_service_proxy() as sp:
                        sp.quit_server()
            except ConnectionError:
                rich.print(f"[red]Couldn't connect to 'tgf4000_cs' {device_id}, process probably not running. ")


@app.command()
def status(
    device_id: Annotated[
        str, typer.Argument(help="Identifies the hardware controller (as per local settings and setup)")
    ],
) -> None:
    """Requests the status information from the Aim-TTi TGF4000 Control Server."""

    hostname = CS_SETTINGS[device_id].get("HOSTNAME", "localhost")
    commanding_port = CS_SETTINGS[device_id].get("COMMANDING_PORT", 0)
    service_port = CS_SETTINGS[device_id].get("SERVICE_PORT", 0)
    monitoring_port = CS_SETTINGS[device_id].get("MONITORING_PORT", 0)
    service_type = CS_SETTINGS[device_id].get("SERVICE_TYPE", "tgf4000_cs")

    if commanding_port != 0:
        endpoint = connect_address(PROTOCOL, hostname, commanding_port)
        port = commanding_port
        service_port = service_port
        monitoring_port = monitoring_port

    else:
        with RegistryClient() as reg:
            service = reg.discover_service(service_type)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]
                service_port = service["metadata"]["service_port"]
                monitoring_port = service["metadata"]["monitoring_port"]
                endpoint = connect_address(protocol, hostname, port)
            else:
                rich.print(
                    f"[red]The Aim-TTi TGF4000 Control Server {device_id} isn't registered as a service. The Control "
                    f"Server cannot be contacted without the required information from the service registry.[/]"
                )
                rich.print(f"Aim-TTi TGF4000 {device_id}: [red]not active")
                return

    # noinspection PyUnboundLocalVariable
    if is_control_server_active(endpoint, timeout=2):
        rich.print(f"Aim-TTi TGF4000 {device_id}: [green]active -> {endpoint}")

        with Tgf4000Proxy(device_id) as tgf4000:
            sim = tgf4000.is_simulator()
            connected = tgf4000.is_connected()
            ip = tgf4000.get_ip_address()
            rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
            rich.print(f"hostname: {ip}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"commanding port: {port}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"service port: {service_port}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"monitoring port: {monitoring_port}")
    else:
        rich.print(f"Aim-TTi TGF4000 {device_id}: [red]not active")


if __name__ == "__main__":
    import logging

    from egse.logger import set_all_logger_levels

    set_all_logger_levels(logging.DEBUG)

    sys.exit(app())



"""
Asynchronous Registry Backend Protocol and Implementations.

This file defines:
1. A Protocol for async registry backends
2. The AsyncSQLiteBackend implementation
3. The AsyncInMemoryBackend for testing
"""

from __future__ import annotations

import asyncio
import json
import time
from contextlib import asynccontextmanager
from typing import Any
from typing import Protocol
from typing import runtime_checkable

import aiosqlite

from egse.decorators import implements_protocol
from egse.registry import logger


@runtime_checkable
class AsyncRegistryBackend(Protocol):
    """Protocol defining the interface for asynchronous registry backends."""

    async def initialize(self) -> None:
        """Initialize the backend."""
        ...

    async def close(self) -> None:
        """Close the backend and release resources."""
        ...

    async def clean_expired_services(self) -> list[str]:
        """
        Remove services that haven't sent a heartbeat within their TTL.

        Returns:
            List of service IDs that were removed.
        """
        ...

    async def register(self, service_id: str, service_info: dict[str, Any], ttl: int = 30) -> bool:
        """
        Register a service with the registry.

        FIXME: describe what is mandatory in the service_info dictionary.

        Args:
            service_id: Unique identifier for the service
            service_info: Service information (host, port, metadata, etc.)
            ttl: Time to live in seconds

        Returns:
            True if registration was successful, False otherwise
        """
        ...

    async def deregister(self, service_id: str) -> bool:
        """
        Remove a service from the registry.

        Args:
            service_id: Unique identifier for the service

        Returns:
            True if de-registration was successful, False otherwise
        """
        ...

    async def renew(self, service_id: str) -> bool:
        """
        Renew a service's TTL in the registry.

        Args:
            service_id: Unique identifier for the service

        Returns:
            True if renewal was successful, False otherwise
        """
        ...

    async def get_service(self, service_id: str) -> dict[str, Any] | None:
        """
        Get information about a specific service.

        Args:
            service_id: Unique identifier for the service

        Returns:
            Service information or None if not found
        """
        ...

    async def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """
        List all registered services, optionally filtered by type.

        Args:
            service_type: Optional service type to filter by

        Returns:
            List of service information dictionaries
        """
        ...

    async def discover_service(self, service_type: str) -> dict[str, Any] | None:
        """
        Find a healthy service of the specified type using load balancing.

        Args:
            service_type: Type of service to discover

        Returns:
            Service information or None if no healthy service is found
        """
        ...


@implements_protocol(AsyncRegistryBackend)
class AsyncSQLiteBackend:
    """Asynchronous persistent storage backend using SQLite."""

    def __init__(self, db_path: str = "service_registry.db"):
        """
        Initialize the SQLite backend.

        Args:
            db_path: Path to the SQLite database file
        """
        self.db_path = db_path
        self.logger = logger
        self._lock = asyncio.Lock()
        self._db = None

    async def initialize(self) -> None:
        """Initialize the database connection and schema."""
        self._db = await aiosqlite.connect(self.db_path)

        # Enable WAL mode for better concurrency
        await self._db.execute("PRAGMA journal_mode=WAL")

        # Initialize schema
        await self._init_db()

        self.logger.info(f"SQLite backend initialized with database: {self.db_path}")

    async def close(self) -> None:
        """Close the database connection."""
        if self._db:
            await self._db.close()
            self._db = None
            self.logger.info("SQLite backend closed")

    @asynccontextmanager
    async def _db_cursor(self):
        """Async context manager for database operations with automatic commit/rollback."""
        async with self._lock:  # Ensure thread safety
            try:
                cursor = await self._db.cursor()
                yield cursor
                await self._db.commit()
            except Exception as exc:
                await self._db.rollback()
                self.logger.error(f"Database error: {exc}")
                raise

    async def _init_db(self) -> None:
        """Initialize the database schema if it doesn't exist."""
        async with self._db_cursor() as cursor:
            # Create services table
            await cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS services (
                    id TEXT PRIMARY KEY,
                    name TEXT NOT NULL,
                    host TEXT NOT NULL,
                    port INTEGER NOT NULL,
                    metadata TEXT,
                    ttl INTEGER NOT NULL,
                    last_heartbeat INTEGER NOT NULL,
                    tags TEXT
                )
                """
            )

    async def clean_expired_services(self) -> list[str]:
        """Remove services that haven't sent a heartbeat within their TTL."""
        current_time = int(time.time())

        async with self._db_cursor() as cursor:
            # Find expired services first (for return value)
            await cursor.execute(
                """
                SELECT id FROM services 
                WHERE last_heartbeat + ttl < ?
                """,
                (current_time,),
            )

            expired_ids = [row[0] async for row in cursor]

            # Then delete them
            if expired_ids:
                await cursor.execute(
                    """
                    DELETE FROM services 
                    WHERE last_heartbeat + ttl < ?
                    """,
                    (current_time,),
                )

                self.logger.info(f"Cleaned up {len(expired_ids)} expired services")

            return expired_ids

    async def register(self, service_id: str, service_info: dict[str, Any], ttl: int = 30) -> bool:
        """Register a service with the registry."""
        try:
            current_time = int(time.time())

            # Prepare data for insertion
            name = service_info.get("name", service_id)
            host = service_info.get("host", "127.0.0.1")
            port = service_info.get("port", 8000)

            metadata = service_info.get("metadata", {}).copy()
            tags = service_info.get("tags", []).copy()
            service_type = service_info.get("type")

            if service_type:
                if service_type not in tags:
                    tags.append(service_type)
                if "type" in metadata:
                    self.logger.warning(
                        f"The 'type' key is found in both 'service_info' ({service_type}) and 'metadata' "
                        f"({metadata['type']}), overwriting in 'metadata."
                    )
                metadata["type"] = service_info["type"]

            # Convert metadata and tags to JSON strings
            metadata_json = json.dumps(metadata)
            tags_json = json.dumps(tags)

            async with self._db_cursor() as cursor:
                # Check if service already exists
                await cursor.execute("SELECT id FROM services WHERE id = ?", (service_id,))
                exists = await cursor.fetchone() is not None

                if exists:
                    # Update existing service
                    await cursor.execute(
                        """
                        UPDATE services
                        SET name = ?, host = ?, port = ?, metadata = ?, 
                            ttl = ?, last_heartbeat = ?, tags = ?
                        WHERE id = ?
                        """,
                        (name, host, port, metadata_json, ttl, current_time, tags_json, service_id),
                    )
                else:
                    # Insert new service
                    await cursor.execute(
                        """
                        INSERT INTO services 
                        (id, name, host, port, metadata, ttl, last_heartbeat, tags)
                        VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                        """,
                        (service_id, name, host, port, metadata_json, ttl, current_time, tags_json),
                    )

            return True
        except Exception as exc:
            self.logger.error(f"Failed to register service {service_id}: {exc}")
            return False

    async def deregister(self, service_id: str) -> bool:
        """Remove a service from the registry."""
        try:
            async with self._db_cursor() as cursor:
                await cursor.execute("DELETE FROM services WHERE id = ?", (service_id,))
                return cursor.rowcount > 0
        except Exception as exc:
            self.logger.error(f"Failed to deregister service {service_id}: {exc}")
            return False

    async def renew(self, service_id: str) -> bool:
        """Renew a service's TTL in the registry."""
        try:
            current_time = int(time.time())

            async with self._db_cursor() as cursor:
                await cursor.execute(
                    """
                    UPDATE services
                    SET last_heartbeat = ?
                    WHERE id = ?
                    """,
                    (current_time, service_id),
                )

                return cursor.rowcount > 0
        except Exception as exc:
            self.logger.error(f"Failed to renew service {service_id}: {exc}")
            return False

    async def get_service(self, service_id: str) -> dict[str, Any] | None:
        """Get information about a specific service."""
        try:
            async with self._db_cursor() as cursor:
                await cursor.execute(
                    """
                    SELECT id, name, host, port, metadata, ttl, last_heartbeat, tags
                    FROM services
                    WHERE id = ?
                    """,
                    (service_id,),
                )

                row = await cursor.fetchone()

                if row:
                    current_time = int(time.time())
                    return {
                        "id": row[0],
                        "name": row[1],
                        "host": row[2],
                        "port": row[3],
                        "metadata": json.loads(row[4]) if row[4] else {},
                        "ttl": row[5],
                        "last_heartbeat": row[6],
                        "tags": json.loads(row[7]) if row[7] else [],
                        "health": "passing" if current_time - row[6] <= row[5] else "critical",
                    }

                return None
        except Exception as exc:
            self.logger.error(f"Failed to get service {service_id}: {exc}")
            return None

    async def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """List all registered services, optionally filtered by type."""
        try:
            async with self._db_cursor() as cursor:
                await cursor.execute(
                    """
                    SELECT id, name, host, port, metadata, ttl, last_heartbeat, tags
                    FROM services
                    """
                )

                services = []
                current_time = int(time.time())

                async for row in cursor:
                    tags = json.loads(row[7]) if row[7] else []
                    metadata = json.loads(row[4]) if row[4] else {}

                    # Check if we need to filter by service type
                    if service_type and service_type not in tags and metadata.get("type") != service_type:
                        continue

                    services.append(
                        {
                            "id": row[0],
                            "name": row[1],
                            "host": row[2],
                            "port": row[3],
                            "metadata": metadata,
                            "ttl": row[5],
                            "last_heartbeat": row[6],
                            "tags": tags,
                            "health": "passing" if current_time - row[6] <= row[5] else "critical",
                        }
                    )

                return services
        except Exception as exc:
            self.logger.error(f"Failed to list services: {exc}")
            return []

    async def discover_service(self, service_type: str) -> dict[str, Any] | None:
        """Find a healthy service of the specified type using load balancing."""
        services = await self.list_services(service_type)

        # Filter only healthy services
        healthy_services = [s for s in services if s.get("health") == "passing"]

        if healthy_services:
            # Simple load balancing - random selection
            import random

            return random.choice(healthy_services)

        return None


@implements_protocol(AsyncRegistryBackend)
class AsyncInMemoryBackend:
    """
    In-memory backend implementation for testing or simple deployments.

    This backend keeps all service information in memory and is not persistent
    between restarts.
    """

    def __init__(self):
        """Initialize the in-memory backend."""
        self.logger = logger
        self._services = {}  # Dictionary to store services
        self._lock = asyncio.Lock()  # Async lock for thread safety

    async def initialize(self) -> None:
        """Initialize the backend (no-op for in-memory)."""
        self.logger.info("In-memory backend initialized")

    async def close(self) -> None:
        """Close the backend (no-op for in-memory)."""
        self.logger.info("In-memory backend closed")

    async def clean_expired_services(self) -> list[str]:
        """Remove services that haven't sent a heartbeat within their TTL.

        Returns:
            A list of the expired and removed id's.
        """
        current_time = int(time.time())
        expired_ids = []

        async with self._lock:
            for service_id, service_data in list(self._services.items()):
                last_heartbeat = service_data.get("last_heartbeat", 0)
                ttl = service_data.get("ttl", 30)

                if current_time - last_heartbeat > ttl:
                    expired_ids.append(service_id)
                    del self._services[service_id]

        if expired_ids:
            self.logger.info(f"Cleaned up {len(expired_ids)} expired services")

        return expired_ids

    async def register(self, service_id: str, service_info: dict[str, Any], ttl: int = 30) -> bool:
        """Register a service with the registry.

        Returns:
            True when the service was properly registered, False is an error occurred.
        """
        try:
            current_time = int(time.time())

            self.logger.info(f"{service_id}: {service_info = }")

            # Make a deepcopy of the service_info
            service_data = json.loads(json.dumps(service_info.copy()))

            # Add TTL and heartbeat information
            service_data["ttl"] = ttl
            service_data["last_heartbeat"] = current_time

            if "metadata" not in service_data:
                service_data["metadata"] = {}
            if "tags" not in service_data:
                service_data["tags"] = []

            # Always add service_type to tags (used for discovery) and to metadata
            service_type = service_data.get("type")
            if service_type:
                if service_type not in service_data["tags"]:
                    service_data["tags"].append(service_type)
                if "type" in service_data["metadata"]:
                    self.logger.warning(
                        f"The 'type' key is found in both 'service_info' ({service_type}) and 'metadata' "
                        f"({service_data['metadata']['type']}), overwriting in 'metadata."
                    )

                service_data["metadata"]["type"] = service_type

            async with self._lock:
                self._services[service_id] = service_data
                self.logger.info(f"{service_id}: {service_data = }")

            return True
        except Exception as exc:
            self.logger.error(f"Failed to register service {service_id}: {exc}")
            return False

    async def deregister(self, service_id: str) -> bool:
        """Remove a service from the registry.

        Returns:
            True when the service was properly de-registered, False when an error
                occurred or when the `service_id` was not found.
        """
        try:
            async with self._lock:
                if service_id in self._services:
                    del self._services[service_id]
                    return True
                return False
        except Exception as exc:
            self.logger.error(f"Failed to deregister service {service_id}: {exc}")
            return False

    async def renew(self, service_id: str) -> bool:
        """Renew a service's TTL in the registry.

        Returns:
            True when the new TTL could be set, False if an error occurred or when
                the `service_id` was not found.
        """
        try:
            current_time = int(time.time())

            async with self._lock:
                if service_id in self._services:
                    self._services[service_id]["last_heartbeat"] = current_time
                    return True
                return False
        except Exception as exc:
            self.logger.error(f"Failed to renew service {service_id}: {exc}")
            return False

    async def get_service(self, service_id: str) -> dict[str, Any] | None:
        """Get information about a specific service.

        Returns:
            A dictionary with information about the service, or None when an
                error occurred or the `service_id` could not be found.
        """
        try:
            async with self._lock:
                if service_id in self._services:
                    service_data = self._services[service_id].copy()

                    # Add the ID to the service data (this was missing!)
                    service_data["id"] = service_id

                    # Add health status
                    current_time = int(time.time())
                    last_heartbeat = service_data.get("last_heartbeat", 0)
                    ttl = service_data.get("ttl", 30)
                    service_data["health"] = "passing" if current_time - last_heartbeat <= ttl else "critical"

                    return service_data
                return None
        except Exception as exc:
            self.logger.error(f"Failed to get service {service_id}: {exc}")
            return None

    async def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """List all registered services, optionally filtered by type.

        Returns:
            A list of dictionaries containing information about the services that
                are registered. An empty list is returned whn no services are registered
                or when an error occurred.
        """
        try:
            services = []
            current_time = int(time.time())

            async with self._lock:
                for service_id, service_data in self._services.items():
                    # Make a copy to avoid modifying the original
                    service = service_data.copy()

                    # Add health status
                    last_heartbeat = service.get("last_heartbeat", 0)
                    ttl = service.get("ttl", 30)
                    service["health"] = "passing" if current_time - last_heartbeat <= ttl else "critical"

                    # Check if we need to filter by service type
                    if service_type:
                        tags = service.get("tags", [])
                        metadata = service.get("metadata", {})

                        if service_type not in tags and metadata.get("type") != service_type:
                            continue

                    services.append(service)

            return services
        except Exception as exc:
            self.logger.error(f"Failed to list services: {exc}")
            return []

    async def discover_service(self, service_type: str) -> dict[str, Any] | None:
        """
        Find a healthy service of the specified type using load balancing.

        Only healthy services will be returned, and if more than one healthy service exists
        for the given service_type, ony of them will be returned by random choice.

        Returns:
            A dictionary with the information for the service. If no healthy service is
                available for the given service type, None will be returned.
        """
        services = await self.list_services(service_type)

        # Filter only healthy services
        healthy_services = [s for s in services if s.get("health") == "passing"]

        if healthy_services:
            # Simple load balancing - pick one of the services by random selection
            # In practice, there will be only one service of a given type, but if you
            # need to get your hands on all services of that type, use `list_services()`.
            import random

            return random.choice(healthy_services)

        return None



"""
Device control for the Symétrie Hexapod PUNA, ZONDA, and JORAN.

This package contains the modules and classes to work with the Hexapod PUNA, the Hexapod ZONDA, and the Hexapod JORAN
from [Symétrie](www.symetrie.fr).

The main entry point for the user of this package is through the terminal commands to start the
control servers for the PUNA, ZONDA, and JORAN Hexapod, and the GUIs that are provided to interact with
the hexapods. The following commands start the control servers for the PUNA, ZONDA, and JORAN
in the background.

    $ puna_cs start-bg
    $ zonda_cs start-bg
    $ joran_cs start-bg

The GUIs can be started with the following commands:

    $ puna_ui
    $ zonda_ui
    $ joran_ui

For developers, the `PunaProxy`, `ZondaProxy`, and `JoranProxy` classes are the main interface to command the
hardware.

For the PUNA:
    >>> from egse.hexapod.symetrie.puna import PunaProxy
    >>> puna = PunaProxy()

For the ZONDA:

    >>> from egse.hexapod.symetrie.zonda import ZondaProxy
    >>> zonda = ZondaProxy()

For the JORAN:

    >>> from egse.hexapod.symetrie.joran import JoranProxy
    >>> joran = JoranProxy()

These classes will connect to their control servers and provide all commands to
control the hexapod and monitor its positions and status.


"""

import logging
from typing import NamedTuple

from egse.device import DeviceFactoryInterface
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.settings import SettingsError

logger = logging.getLogger("egse.hexapod.symetrie")

HEXAPOD_SETTINGS = Settings.load("Hexapod Controller")

HexapodInfo = NamedTuple(
    "HexapodInfo",
    [
        ("hostname", str),
        ("port", int),
        ("device_id", str),
        ("device_name", str),
        ("device_type", str),
        ("controller_type", str),
    ],
)


def get_hexapod_controller_pars(device_id: str) -> HexapodInfo:
    """
    Returns a NamedTuple HexapodInfo with the hostname (str), port number (int),
    device id (str), device name (str), device_type (str) and controller type (str)
    for the hexapod controller as defined in the Settings.

    Note the returned values are for the device hardware controller, not the control server.

    All values are derived from the Settings.
    """

    logger.debug(f"Getting parameters for {device_id} controller...")

    try:
        hostname: str = HEXAPOD_SETTINGS[device_id]["HOSTNAME"]
        port: int = int(HEXAPOD_SETTINGS[device_id]["PORT"])
        controller_type: str = HEXAPOD_SETTINGS[device_id]["CONTROLLER_TYPE"]
        device_name: str = HEXAPOD_SETTINGS[device_id]["DEVICE_NAME"]
        device_type: str = HEXAPOD_SETTINGS[device_id]["DEVICE_TYPE"]
    except (KeyError, AttributeError) as exc:
        raise SettingsError("The Settings do not contain proper controller parameters for the Hexapod.") from exc

    logger.debug(f"{hostname=}, {port=}, {device_id=}, {device_name=}, {device_type=}, {controller_type=}")
    return HexapodInfo(hostname, port, device_id, device_name, device_type, controller_type)


class ProxyFactory(DeviceFactoryInterface):
    """
    A factory class that will create the Proxy that matches the given device name and identifier.

    The device name is matched against the string 'puna' or 'zonda'. If the device name doesn't contain
    one of these names, a ValueError will be raised.
    """

    def create(self, device_type: str, *, device_id: str = None, **_ignored):
        logger.debug(f"{device_type=}, {device_id=}")

        with RegistryClient() as reg:
            service = reg.discover_service(device_id)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]

            else:
                raise RuntimeError(f"No service registered as {device_id}")

        if "puna" in device_type.lower():
            controller_type = HEXAPOD_SETTINGS[device_id]["CONTROLLER_TYPE"]
            if controller_type.lower() == "alpha":
                from egse.hexapod.symetrie.puna import PunaProxy

                return PunaProxy(protocol, hostname, port)
            elif controller_type == "alpha_plus":
                from egse.hexapod.symetrie.punaplus import PunaPlusProxy

                return PunaPlusProxy(protocol, hostname, port)
            else:
                raise ValueError(f"Unknown controller_type ({controller_type}) for {device_type} – {device_id}")

        elif "zonda" in device_type.lower():
            from egse.hexapod.symetrie.zonda import ZondaProxy

            return ZondaProxy(protocol, hostname, port)

        elif "joran" in device_type.lower():
            from egse.hexapod.symetrie.joran import JoranProxy

            return JoranProxy(protocol, hostname, port)

        else:
            raise ValueError(f"Unknown device type: {device_type}")


class ControllerFactory(DeviceFactoryInterface):
    """
    A factory class that will create the Controller that matches the given device name and identifier.

    The device name is matched against the string 'puna', 'zonda', or 'joran'. If the device name doesn't contain
    one of these names, a ValueError will be raised.
    """

    def create(self, device_type: str, *, device_id: str = None, **_ignored):
        if "puna" in device_type.lower():
            from egse.hexapod.symetrie.puna import PunaController
            from egse.hexapod.symetrie.punaplus import PunaPlusController

            hostname = HEXAPOD_SETTINGS[device_id]["HOSTNAME"]
            port = HEXAPOD_SETTINGS[device_id]["PORT"]
            controller_type = HEXAPOD_SETTINGS[device_id]["CONTROLLER_TYPE"]
            if controller_type.lower() == "alpha":
                return PunaController(hostname=hostname, port=port)
            elif controller_type.lower() == "alpha_plus":
                return PunaPlusController(hostname=hostname, port=port)
            else:
                raise ValueError(f"Unknown controller_type ({controller_type}) for {device_type} – {device_id}")

        elif "zonda" in device_type.lower():
            from egse.hexapod.symetrie.zonda import ZondaController

            hostname = HEXAPOD_SETTINGS[device_id]["HOSTNAME"]
            port = HEXAPOD_SETTINGS[device_id]["PORT"]

            return ZondaController(hostname=hostname, port=port)

        elif "joran" in device_type.lower():
            from egse.hexapod.symetrie.joran import JoranController

            return JoranController()

        else:
            raise ValueError(f"Unknown device name: {device_type}")



import math
import numpy as np

from egse.bits import set_bit
from egse.coordinates.reference_frame import ReferenceFrame
from egse.hexapod.symetrie import logger
from egse.hexapod.symetrie.pmac import decode_Q36


class HexapodSimulator:
    """
    HexapodSimulator simulates the Symétrie Hexapod. The class is heavily based on the
    ReferenceFrames in the `egse.coordinates` package.

    The simulator implements the same methods as the HexapodController class which acts on the
    real hardware controller in either simulation mode or with a real Hexapod PUNA connected.

    Therefore, the HexapodSimulator can be used instead of the Hexapod class in test harnesses
    and when the hardware is not available.

    This class simulates all the movements and status of the Hexapod.
    """

    def __init__(self):
        identity = np.identity(4)

        # Rotation around static axis, and around x, y and z in that order
        self.rot_config = "sxyz"

        # Configure the Master Reference Frame
        self.cs_master = ReferenceFrame.create_master()

        # Configure the Machine Coordinate System, i.e. cs_mec [ref:cs_master]
        self.cs_machine = ReferenceFrame(
            transformation=identity,
            ref=self.cs_master,
            name="Machine[Master]",
            rot_config=self.rot_config,
        )

        # Configure the Platform Coordinate System, i.e. cs_platform [ref:cs_machine]
        # default after homing: PLATFORM = MACHINE

        self.cs_platform = ReferenceFrame(
            transformation=identity,
            ref=self.cs_machine,
            name="Platform[Machine]",
            rot_config=self.rot_config,
        )

        # Configure the User Coordinate System, i.e. cs_user [ref:cs_machine]
        self.cs_user = ReferenceFrame(
            transformation=identity,
            ref=self.cs_machine,
            name="User[Machine]",
            rot_config=self.rot_config,
        )

        # Configure the Object Coordinate System, i.e. cs_object [ref:cs_platform]
        self.cs_object = ReferenceFrame(
            transformation=identity,
            ref=self.cs_platform,
            name="Object[Platform]",
            rot_config=self.rot_config,
        )

        # We use a CS called cs_object_in_user, i.e. Object as defined in the User CS,
        # and we define this
        # from the transformation user -> object.

        tf_user_to_object = self.cs_user.get_active_transformation_to(self.cs_object)
        self.cs_object_in_user = ReferenceFrame(
            tf_user_to_object, rot_config=self.rot_config, ref=self.cs_user, name="Object[User]"
        )

        # Define the invariant links within the system, i.e. some systems are bound with an
        # invariant transformation
        # matrix and those links shall be preserved throughout the movement within the system.

        # We link this cs_object_in_user to cs_object with the identity transformation,
        # which connects them together

        self.cs_object_in_user.add_link(self.cs_object, transformation=identity)

        # The User Coordinate System is linked to the Machine Coordinate System

        self.cs_machine.add_link(self.cs_user, transformation=self.cs_user.transformation)

        # The Object Coordinate System is linked to the Platform Coordinate System

        self.cs_platform.add_link(self.cs_object, transformation=self.cs_object.transformation)

        # Keep a record if the homing() command has been executed.

        self.homing_done = False
        self.control_loop = False
        self._virtual_homing = False
        self._virtual_homing_position = None

        # Just keep the speed settings, no used in movement currently

        self._speed = [1.0, 1.0, 0.01, 0.001, 4.0, 2.0]

        # Print out some debugging information

        logger.debug(f"Linked to cs_object_in_user  {[i.name for i in self.cs_object_in_user.linked_to]}")
        logger.debug(f"Linked to cs_object          {[i.name for i in self.cs_object.linked_to]}")
        logger.debug(f"Linked to cs_platform        {[i.name for i in self.cs_platform.linked_to]}")
        logger.debug(f"Linked to cs_machine         {[i.name for i in self.cs_machine.linked_to or {}]}")

    def is_simulator(self):
        return True

    def connect(self):
        pass

    def reconnect(self):
        pass

    def disconnect(self):
        # TODO:
        #   Should I keep state in this class to check if it has been disconnected?
        #
        # TODO:
        #   What happens when I re-connect to this Simulator? Shall it be in Homing position or
        #   do I have to keep state via a persistency mechanism?
        pass

    def is_connected(self):
        return True

    def reset(self, wait=True, verbose=False):
        # TODO:
        #   Find out what exactly a reset() should be doing. Does it bring back the Hexapod
        #   in it's original state, loosing all definitions of coordinate systems? Or does it
        #   do a clearError() and a homing()?
        pass

    def homing(self):
        self.goto_zero_position()
        self.homing_done = True
        self._virtual_homing = False
        self._virtual_homing_position = None
        return 0

    def is_homing_done(self):
        return self.homing_done

    def set_virtual_homing(self, tx, ty, tz, rx, ry, rz):
        self._virtual_homing_position = [tx, ty, tz, rx, ry, rz]
        self._virtual_homing = True
        return 0

    def stop(self):
        pass

    def clear_error(self):
        return 0

    def activate_control_loop(self):
        self.control_loop = True
        return self.control_loop

    def deactivate_control_loop(self):
        self.control_loop = False
        return self.control_loop

    def configure_coordinates_systems(self, tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o):
        identity = np.identity(4)

        # Redefine the User Coordinate System

        translation = np.array([tx_u, ty_u, tz_u])
        rotation = np.array([rx_u, ry_u, rz_u])
        degrees = True

        # Remove the old links between user and machine CS, and between Object in User and Object CS

        self.cs_machine.remove_link(self.cs_user)
        self.cs_object_in_user.remove_link(self.cs_object)

        # Redefine the User Coordinate System

        self.cs_user = ReferenceFrame.from_translation_rotation(
            translation,
            rotation,
            rotation_config=self.rot_config,
            reference_frame=self.cs_machine,
            name="User[Machine]",
            degrees=degrees,
        )

        # Redefine the Object in User Coordinate System

        tf_user_to_object = self.cs_user.get_active_transformation_to(self.cs_object)
        self.cs_object_in_user = ReferenceFrame(
            tf_user_to_object, rot_config=self.rot_config, ref=self.cs_user, name="Object[User]"
        )

        # Define the invariant links within the system, i.e. some systems are bound with an
        # invariant transformation
        # matrix and those links shall be preserved throughout the movement within the system.

        # User and Machine CS are invariant, reset the transformation. User in Object is
        # identical to Object

        self.cs_machine.add_link(self.cs_user, transformation=self.cs_user.transformation)
        self.cs_object_in_user.add_link(self.cs_object, transformation=identity)

        # Redefine the Object Coordinates System

        translation = np.array([tx_o, ty_o, tz_o])
        rotation = np.array([rx_o, ry_o, rz_o])
        degrees = True

        # Remove the old links between user and machine CS, and between Object in User and Object CS

        self.cs_platform.remove_link(self.cs_object)
        self.cs_object_in_user.remove_link(self.cs_object)

        self.cs_object = ReferenceFrame.from_translation_rotation(
            translation,
            rotation,
            rotation_config=self.rot_config,
            reference_frame=self.cs_platform,
            name="Object[Platform]",
            degrees=degrees,
        )

        # Redefine the Object in User Coordinate System

        tf_user_to_object = self.cs_user.get_active_transformation_to(self.cs_object)
        self.cs_object_in_user = ReferenceFrame(
            tf_user_to_object, rot_config=self.rot_config, ref=self.cs_user, name="Object[User]"
        )

        # Object CS and Platform CS are invariant, reset the transformation. User in Object is
        # identical to Object

        self.cs_platform.add_link(self.cs_object, transformation=self.cs_object.transformation)
        self.cs_object_in_user.add_link(self.cs_object, transformation=identity)

        return 0

    def get_coordinates_systems(self):
        degrees = True

        t_user, r_user = self.cs_user.get_translation_rotation_vectors(degrees=degrees)
        t_object, r_object = self.cs_object.get_translation_rotation_vectors(degrees=degrees)

        return list(np.concatenate((t_user, r_user, t_object, r_object)))

    def move_absolute(self, tx, ty, tz, rx, ry, rz):
        # FIXME:
        #  to really simulate the behavior of the Hexapod, this method should implement limit
        #  checking and other condition or error checking, e.g. argument matching, etc.

        logger.debug(f"moveAbsolute with {tx}, {ty}, {tz}, {rx}, {ry}, {rz}")

        translation = np.array([tx, ty, tz])
        rotation = np.array([rx, ry, rz])

        # We set a new transformation for cs_object_in_user which will update our model,
        # because cs_object and cs_object_in_user are linked.

        self.cs_object_in_user.set_translation_rotation(
            translation,
            rotation,
            rotation_config=self.rot_config,
            active=True,
            degrees=True,
            preserve_links=True,
        )

        return 0

    def move_relative_object(self, tx, ty, tz, rx, ry, rz):
        tr_rel = np.array([tx, ty, tz])
        rot_rel = np.array([rx, ry, rz])

        self.cs_object.apply_translation_rotation(
            tr_rel,
            rot_rel,
            rotation_config=self.rot_config,
            active=True,
            degrees=True,
            preserve_links=True,
        )

    def move_relative_user(self, tx, ty, tz, rx, ry, rz):
        # The Symétrie Hexapod definition of moveRelativeUser
        #
        # - Translation and rotations are expressed in USER reference frame
        #
        # - Actually,
        #     - the translations happen parallel to the USER reference frame
        #     - the rotations are applied after the translations, on the OBJ ReferenceFrame
        #
        # To achieve this,
        #
        #     * OBUSR is "de-rotated", to become parallel to USR
        #     * the requested translation is applied
        #     * OBUSR is "re-rotated" to its original orientation
        #     * the requested rotation is applied

        # Derotation of cs_object --> cs_user

        derotation = self.cs_object.get_active_transformation_to(self.cs_user)
        derotation[:3, 3] = [0, 0, 0]

        # Reverse rotation

        rerotation = derotation.T

        # Requested translation matrix

        translation = np.identity(4)
        translation[:3, 3] = [tx, ty, tz]

        # Requested rotation matrix

        import egse.coordinates.transform3d_addon as t3add

        rotation = t3add.translation_rotation_to_transformation(
            [0, 0, 0], [rx, ry, rz], rotation_config=self.rot_config
        )

        transformation = derotation @ translation @ rerotation @ rotation

        # Adapt our model

        self.cs_object.apply_transformation(transformation)

    def check_absolute_movement(self, tx, ty, tz, rx, ry, rz):
        rc = 0
        rc_dict = {}
        if not -30.0 <= tx <= 30.0:
            rc += 1
            rc_dict.update({1: "Tx should be in range ±30.0 mm"})
        if not -30.0 <= ty <= 30.0:
            rc += 1
            rc_dict.update({2: "Ty should be in range ±30.0 mm"})
        if not -20.0 <= tz <= 20.0:
            rc += 1
            rc_dict.update({3: "Tz should be in range ±20.0 mm"})
        if not -11.0 <= rx <= 11.0:
            rc += 1
            rc_dict.update({4: "Rx should be in range ±11.0 mm"})
        if not -11.0 <= ry <= 11.0:
            rc += 1
            rc_dict.update({5: "Ry should be in range ±11.0 mm"})
        if not -20.0 <= rz <= 20.0:
            rc += 1
            rc_dict.update({6: "Rz should be in range ±20.0 mm"})
        return rc, rc_dict

    def check_relative_object_movement(self, tx, ty, tz, rx, ry, rz):
        return 0, {}

    def check_relative_user_movement(self, tx, ty, tz, rx, ry, rz):
        return 0, {}

    def get_user_positions(self):
        t, r = self.cs_user.get_active_translation_rotation_vectors_to(self.cs_object_in_user)

        pos = list(np.concatenate((t, r)))

        return pos

    def get_machine_positions(self):
        t, r = self.cs_platform.get_translation_rotation_vectors()
        t, r = self.cs_machine.get_active_translation_rotation_vectors_to(self.cs_platform)

        pos = list(np.concatenate((t, r)))

        return pos

    def get_actuator_length(self):
        alen = [math.nan for _ in range(6)]

        return alen

    def get_general_state(self):
        state = 0
        state = set_bit(state, 1)  # System Initialized
        state = set_bit(state, 2)  # In Position
        if self.homing_done:
            state = set_bit(state, 4)
        if self.control_loop:
            state = set_bit(state, 3)
        if self._virtual_homing:
            state = set_bit(state, 18)
        return state, decode_Q36(state)

    def goto_specific_position(self, pos):
        return 0

    def goto_retracted_position(self):
        translation = np.array([0, 0, -20])
        rotation = np.array([0, 0, 0])

        self.cs_platform.set_translation_rotation(
            translation,
            rotation,
            rotation_config=self.rot_config,
            active=True,
            degrees=True,
            preserve_links=True,
        )

        return 0

    def goto_zero_position(self):
        # We set a new transformation for cs_platform which will update our model.
        # See issue #58: updating the cs_platform is currently not a good idea because that will
        # not properly update the chain/path upwards, i.e. the chain/path is followed in the
        # direction of the references that are defined, not in the other direction.
        #
        translation = np.array([0, 0, 0])
        rotation = np.array([0, 0, 0])

        self.cs_platform.set_translation_rotation(
            translation,
            rotation,
            rotation_config=self.rot_config,
            active=True,
            degrees=True,
            preserve_links=True,
        )

        # As a work around for the bug in issue #58, we determine the transformation from
        # cs_object as it is
        # invariantly linked with cs_platform. Updating cs_object_in_user with that
        # transformation will
        # properly update all the reference frames.

        # tr_abs, rot_abs = self.cs_object.getTranslationRotationVectors()
        #
        # self.cs_object_in_user.setTranslationRotation(
        #     tr_abs,
        #     rot_abs,
        #     rot_config=self.rot_config,
        #     active=True,
        #     degrees=True,
        #     preserveLinks=True,
        # )

        return 0

    def is_in_position(self):
        return True

    def jog(self, axis: int, inc: float) -> int:
        pass

    def get_debug_info(self):
        pass

    def set_speed(self, vt, vr):
        self._speed[0] = vt
        self._speed[1] = vr

    def get_speed(self):
        return tuple(self._speed)

    def get_actuator_state(self):
        return [
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
            (
                {
                    0: "In position",
                    1: "Control loop on servo motors active",
                    2: "Homing done",
                    4: 'Input "Positive limit switch"',
                    5: 'Input "Negative limit switch"',
                    6: "Brake control output",
                },
                [1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ),
        ]

    def perform_maintenance(self, axis):
        pass

    def info(self):
        msg = "Info about the PunaSimulator:\n"
        msg += "\n"
        msg += "This Hexapod PUNA Simulator works with several reference frames:\n"
        msg += "  * The machine reference frame\n"
        msg += "  * The platform reference frame\n"
        msg += "  * The object reference frame\n"
        msg += "  * The user reference frame\n\n"
        msg += "Any movement commands result in a transformation of the appropriate coordinate systems."

        logger.info(msg)

        return msg
